from __future__ import annotations

import contextlib
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from graphene import Argument, Boolean, DateTime, Field, InputObjectType, Int, List, NonNull, ObjectType, String
from graphene import Enum as GrapheneEnum
from opentelemetry import trace

from infrahub.core import registry
from infrahub.core.constants import DiffAction, RelationshipCardinality, RelationshipDirection
from infrahub.core.constants.database import DatabaseEdgeType
from infrahub.core.diff.diff_locker import DiffLocker
from infrahub.core.diff.model.path import BranchTrackingId, NameTrackingId, TrackingId
from infrahub.core.diff.query.filters import EnrichedDiffQueryFilters
from infrahub.core.diff.repository.repository import DiffRepository
from infrahub.core.query.diff import DiffCountChanges
from infrahub.core.timestamp import Timestamp
from infrahub.dependencies.registry import get_component_registry
from infrahub.exceptions import BranchNotFoundError, SchemaNotFoundError, ValidationError
from infrahub.graphql.enums import ConflictSelection as GraphQLConflictSelection
from infrahub.graphql.field_extractor import extract_graphql_fields

if TYPE_CHECKING:
    from datetime import datetime

    from graphql import GraphQLResolveInfo

    from infrahub.core.diff.model.path import (
        EnrichedDiffAttribute,
        EnrichedDiffConflict,
        EnrichedDiffNode,
        EnrichedDiffProperty,
        EnrichedDiffRelationship,
        EnrichedDiffRoot,
        EnrichedDiffSingleRelationship,
    )
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.database import InfrahubDatabase
    from infrahub.graphql.initialization import GraphqlContext

GrapheneDiffActionEnum = GrapheneEnum.from_enum(DiffAction)
GrapheneCardinalityEnum = GrapheneEnum.from_enum(RelationshipCardinality)


@dataclass
class DiffBranchInfo:
    name: str
    branch_start_timestamp: Timestamp | None
    is_terminal: bool


@dataclass
class DiffQueryParams:
    from_time: Timestamp | None
    to_time: Timestamp | None
    tracking_id: TrackingId | None
    exclude_merged: bool


@dataclass
class ParentNodeInfo:
    node: EnrichedDiffNode
    relationship_name: str


class ConflictDetails(ObjectType):
    uuid = String(required=True)
    base_branch_action = Field(GrapheneDiffActionEnum, required=True)
    base_branch_value = String()
    base_branch_changed_at = DateTime(required=True)
    base_branch_label = String()
    diff_branch_action = Field(GrapheneDiffActionEnum, required=True)
    diff_branch_value = String()
    diff_branch_changed_at = DateTime(required=True)
    diff_branch_label = String()
    selected_branch = Field(GraphQLConflictSelection)
    resolvable = Boolean()


class DiffSummaryCounts(ObjectType):
    num_added = Int(required=True)
    num_updated = Int(required=True)
    num_removed = Int(required=True)
    num_conflicts = Int(required=True)


class DiffProperty(ObjectType):
    property_type = String(required=True)
    last_changed_at = DateTime(required=True)
    previous_value = String(required=False)
    new_value = String(required=False)
    previous_label = String(required=False)
    new_label = String(required=False)
    status = Field(GrapheneDiffActionEnum, required=True)
    path_identifier = String(required=True)
    conflict = Field(ConflictDetails, required=False)


class DiffAttribute(DiffSummaryCounts):
    name = String(required=True)
    last_changed_at = DateTime(required=True)
    status = Field(GrapheneDiffActionEnum, required=True)
    path_identifier = String(required=True)
    properties = List(NonNull(DiffProperty))
    contains_conflict = Boolean(required=True)
    conflict = Field(ConflictDetails, required=False)


class DiffSingleRelationship(DiffSummaryCounts):
    last_changed_at = DateTime(required=False)
    status = Field(GrapheneDiffActionEnum, required=True)
    peer_id = String(required=True)
    peer_label = String(required=False)
    path_identifier = String(required=True)
    contains_conflict = Boolean(required=True)
    conflict = Field(ConflictDetails, required=False)
    properties = List(NonNull(DiffProperty))


class DiffRelationship(DiffSummaryCounts):
    name = String(required=True)
    label = String(required=False)
    last_changed_at = DateTime(required=False)
    cardinality = Field(GrapheneCardinalityEnum, required=True)
    status = Field(GrapheneDiffActionEnum, required=True)
    path_identifier = String(required=True)
    elements = List(NonNull(DiffSingleRelationship), required=True)
    contains_conflict = Boolean(required=True)


class DiffNodeParent(ObjectType):
    uuid = String(required=True)
    kind = String(required=False)
    relationship_name = String(required=False)


class DiffNode(DiffSummaryCounts):
    uuid = String(required=True)
    kind = String(required=True)
    label = String(required=True)
    status = Field(GrapheneDiffActionEnum, required=True)
    path_identifier = String(required=True)
    conflict = Field(ConflictDetails, required=False)
    contains_conflict = Boolean(required=True)
    last_changed_at = DateTime(required=False)
    parent = Field(DiffNodeParent, required=False)
    attributes = List(NonNull(DiffAttribute), required=True)
    relationships = List(NonNull(DiffRelationship), required=True)


class DiffTree(DiffSummaryCounts):
    base_branch = String(required=True)
    diff_branch = String(required=True)
    from_time = DateTime(required=True)
    to_time = DateTime(required=True)
    num_untracked_base_changes = Int(required=False)
    num_untracked_diff_changes = Int(required=False)
    name = String(required=False)
    nodes = List(NonNull(DiffNode))


class DiffTreeSummary(DiffSummaryCounts):
    base_branch = String(required=True)
    diff_branch = String(required=True)
    from_time = DateTime(required=True)
    to_time = DateTime(required=True)
    num_unchanged = Int(required=True)
    num_untracked_base_changes = Int(required=False)
    num_untracked_diff_changes = Int(required=False)


class DiffTreeResolver:
    def __init__(self) -> None:
        self.source_branch_name: str | None = None

    def initialize(self, enriched_diff_root: EnrichedDiffRoot) -> None:
        self.source_branch_name = enriched_diff_root.diff_branch_name

    async def to_diff_tree(
        self, enriched_diff_root: EnrichedDiffRoot, graphql_context: GraphqlContext | None = None
    ) -> DiffTree:
        self.initialize(enriched_diff_root=enriched_diff_root)
        all_nodes = list(enriched_diff_root.nodes)
        tree_nodes = [self.to_diff_node(enriched_node=e_node, graphql_context=graphql_context) for e_node in all_nodes]
        name = None
        if enriched_diff_root.tracking_id and isinstance(enriched_diff_root.tracking_id, NameTrackingId):
            name = enriched_diff_root.tracking_id.name
        return DiffTree(
            base_branch=enriched_diff_root.base_branch_name,
            diff_branch=enriched_diff_root.diff_branch_name,
            from_time=await enriched_diff_root.from_time.to_graphql(),
            to_time=await enriched_diff_root.to_time.to_graphql(),
            name=name,
            nodes=tree_nodes,
            num_added=enriched_diff_root.num_added,
            num_updated=enriched_diff_root.num_updated,
            num_removed=enriched_diff_root.num_removed,
            num_conflicts=enriched_diff_root.num_conflicts,
        )

    def _get_parent_info(
        self, diff_node: EnrichedDiffNode, graphql_context: GraphqlContext | None = None
    ) -> ParentNodeInfo | None:
        for r in diff_node.relationships:
            for n in r.nodes:
                relationship_name: str = "undefined"

                if not graphql_context or not self.source_branch_name:
                    return ParentNodeInfo(node=n, relationship_name=relationship_name)

                node_schema: MainSchemaTypes | None = None
                parent_schema: MainSchemaTypes | None = None
                for branch_name in [self.source_branch_name, registry.default_branch]:
                    with contextlib.suppress(BranchNotFoundError, SchemaNotFoundError):
                        node_schema = graphql_context.db.schema.get(
                            name=diff_node.kind, branch=branch_name, duplicate=False
                        )
                        parent_schema = graphql_context.db.schema.get(name=n.kind, branch=branch_name, duplicate=False)
                        break

                if not node_schema or not parent_schema:
                    return ParentNodeInfo(node=n, relationship_name=relationship_name)

                rel_schema = node_schema.get_relationship(name=r.name)
                rels_parent = parent_schema.get_relationships_by_identifier(id=rel_schema.get_identifier())

                if rels_parent and len(rels_parent) == 1:
                    relationship_name = rels_parent[0].name
                elif rels_parent and len(rels_parent) > 1:
                    for rel_parent in rels_parent:
                        if (
                            rel_schema.direction == RelationshipDirection.INBOUND
                            and rel_parent.direction == RelationshipDirection.OUTBOUND
                        ) or (
                            rel_schema.direction == RelationshipDirection.OUTBOUND
                            and rel_parent.direction == RelationshipDirection.INBOUND
                        ):
                            relationship_name = rel_parent.name
                            break

                return ParentNodeInfo(node=n, relationship_name=relationship_name)
        return None

    def to_diff_node(self, enriched_node: EnrichedDiffNode, graphql_context: GraphqlContext | None = None) -> DiffNode:
        diff_attributes = [
            self.to_diff_attribute(enriched_attribute=e_attr, graphql_context=graphql_context)
            for e_attr in enriched_node.attributes
        ]
        diff_relationships = [
            self.to_diff_relationship(enriched_relationship=e_rel, graphql_context=graphql_context)
            for e_rel in enriched_node.relationships
            if e_rel.include_in_response
        ]
        conflict = None
        if enriched_node.conflict:
            conflict = self.to_diff_conflict(enriched_conflict=enriched_node.conflict, graphql_context=graphql_context)

        parent = None
        if parent_info := self._get_parent_info(diff_node=enriched_node, graphql_context=graphql_context):
            parent = DiffNodeParent(
                uuid=parent_info.node.uuid,
                kind=parent_info.node.kind,
                relationship_name=parent_info.relationship_name,
            )

        return DiffNode(
            uuid=enriched_node.uuid,
            kind=enriched_node.kind,
            label=enriched_node.label,
            status=enriched_node.action,
            parent=parent,
            last_changed_at=enriched_node.changed_at.to_datetime() if enriched_node.changed_at else None,
            path_identifier=enriched_node.path_identifier,
            attributes=diff_attributes,
            relationships=diff_relationships,
            contains_conflict=enriched_node.contains_conflict,
            conflict=conflict,
            num_added=enriched_node.num_added,
            num_updated=enriched_node.num_updated,
            num_removed=enriched_node.num_removed,
            num_conflicts=enriched_node.num_conflicts,
        )

    def to_diff_attribute(
        self, enriched_attribute: EnrichedDiffAttribute, graphql_context: GraphqlContext | None = None
    ) -> DiffAttribute:
        diff_properties = [
            self.to_diff_property(enriched_property=e_prop, graphql_context=graphql_context)
            for e_prop in enriched_attribute.properties
        ]
        conflict = None
        for diff_prop in diff_properties:
            if diff_prop.property_type == DatabaseEdgeType.HAS_VALUE.value and diff_prop.conflict:
                conflict = diff_prop.conflict
                diff_prop.conflict = None
        return DiffAttribute(
            name=enriched_attribute.name,
            last_changed_at=enriched_attribute.changed_at.to_datetime(),
            status=enriched_attribute.action,
            path_identifier=enriched_attribute.path_identifier,
            properties=diff_properties,
            contains_conflict=enriched_attribute.contains_conflict,
            conflict=conflict,
            num_added=enriched_attribute.num_added,
            num_updated=enriched_attribute.num_updated,
            num_removed=enriched_attribute.num_removed,
            num_conflicts=enriched_attribute.num_conflicts,
        )

    def to_diff_relationship(
        self, enriched_relationship: EnrichedDiffRelationship, graphql_context: GraphqlContext | None = None
    ) -> DiffRelationship:
        diff_elements = [
            self.to_diff_relationship_element(enriched_element=element, graphql_context=graphql_context)
            for element in enriched_relationship.relationships
        ]
        return DiffRelationship(
            name=enriched_relationship.name,
            label=enriched_relationship.label,
            last_changed_at=enriched_relationship.changed_at.to_datetime()
            if enriched_relationship.changed_at
            else None,
            status=enriched_relationship.action,
            cardinality=enriched_relationship.cardinality,
            path_identifier=enriched_relationship.path_identifier,
            elements=diff_elements,
            contains_conflict=enriched_relationship.contains_conflict,
            num_added=enriched_relationship.num_added,
            num_updated=enriched_relationship.num_updated,
            num_removed=enriched_relationship.num_removed,
            num_conflicts=enriched_relationship.num_conflicts,
        )

    def to_diff_relationship_element(
        self, enriched_element: EnrichedDiffSingleRelationship, graphql_context: GraphqlContext | None = None
    ) -> DiffSingleRelationship:
        diff_properties = [self.to_diff_property(e_prop) for e_prop in enriched_element.properties]
        conflict = None
        if enriched_element.conflict:
            conflict = self.to_diff_conflict(
                enriched_conflict=enriched_element.conflict, graphql_context=graphql_context
            )
        return DiffSingleRelationship(
            last_changed_at=enriched_element.changed_at.to_datetime(),
            status=enriched_element.action,
            peer_id=enriched_element.peer_id,
            peer_label=enriched_element.peer_label,
            path_identifier=enriched_element.path_identifier,
            conflict=conflict,
            properties=diff_properties,
            contains_conflict=enriched_element.contains_conflict,
            num_added=enriched_element.num_added,
            num_updated=enriched_element.num_updated,
            num_removed=enriched_element.num_removed,
            num_conflicts=enriched_element.num_conflicts,
        )

    def to_diff_property(
        self, enriched_property: EnrichedDiffProperty, graphql_context: GraphqlContext | None = None
    ) -> DiffProperty:
        conflict = None
        if enriched_property.conflict:
            conflict = self.to_diff_conflict(
                enriched_conflict=enriched_property.conflict, graphql_context=graphql_context
            )
        return DiffProperty(
            property_type=enriched_property.property_type.value,
            last_changed_at=enriched_property.changed_at.to_datetime(),
            previous_value=enriched_property.previous_value,
            new_value=enriched_property.new_value,
            previous_label=enriched_property.previous_label,
            new_label=enriched_property.new_label,
            status=enriched_property.action,
            path_identifier=enriched_property.path_identifier,
            conflict=conflict,
        )

    def to_diff_conflict(
        self,
        enriched_conflict: EnrichedDiffConflict,
        graphql_context: GraphqlContext | None = None,  # noqa: ARG002
    ) -> ConflictDetails:
        return ConflictDetails(
            uuid=enriched_conflict.uuid,
            base_branch_action=enriched_conflict.base_branch_action,
            base_branch_value=enriched_conflict.base_branch_value,
            base_branch_changed_at=enriched_conflict.base_branch_changed_at.to_datetime()
            if enriched_conflict.base_branch_changed_at
            else None,
            base_branch_label=enriched_conflict.base_branch_label,
            diff_branch_action=enriched_conflict.diff_branch_action,
            diff_branch_value=enriched_conflict.diff_branch_value,
            diff_branch_changed_at=enriched_conflict.diff_branch_changed_at.to_datetime()
            if enriched_conflict.diff_branch_changed_at
            else None,
            diff_branch_label=enriched_conflict.diff_branch_label,
            selected_branch=enriched_conflict.selected_branch.value if enriched_conflict.selected_branch else None,
            resolvable=enriched_conflict.resolvable,
        )

    async def to_graphql(
        self, fields: dict[str, dict], diff_object: Any | None
    ) -> list[dict[str, Any]] | dict[str, Any] | None:
        if diff_object is None:
            return None
        if isinstance(diff_object, list):
            list_response = True
            diff_elements = diff_object
        else:
            list_response = False
            diff_elements = [diff_object]

        response_list = []
        for diff_object_element in diff_elements:
            element_response = {}
            for field_name, sub_fields in fields.items():
                if sub_fields is None:
                    element_response[field_name] = getattr(diff_object_element, field_name, None)
                elif hasattr(diff_object_element, field_name):
                    element_response[field_name] = await self.to_graphql(
                        sub_fields, getattr(diff_object_element, field_name)
                    )
                else:
                    continue
            response_list.append(element_response)
        if list_response:
            return response_list
        return response_list[0]

    async def _add_untracked_fields(
        self,
        db: InfrahubDatabase,
        diff_response: DiffTreeSummary | DiffTree,
        from_time: Timestamp,
        base_branch_name: str | None = None,
        diff_branch_name: str | None = None,
    ) -> None:
        if not (base_branch_name or diff_branch_name):
            return
        branch_names = []
        if base_branch_name:
            branch_names.append(base_branch_name)
        if diff_branch_name:
            branch_names.append(diff_branch_name)
        query = await DiffCountChanges.init(db=db, branch_names=branch_names, diff_from=from_time, diff_to=Timestamp())
        await query.execute(db=db)
        branch_change_map = query.get_num_changes_by_branch()
        if base_branch_name:
            diff_response.num_untracked_base_changes = branch_change_map.get(base_branch_name, 0)
        if diff_branch_name:
            diff_response.num_untracked_diff_changes = branch_change_map.get(diff_branch_name, 0)

    def _get_timestamp(
        self,
        from_time: datetime | None,
        to_time: datetime | None,
        branch_start_timestamp: Timestamp | None,
        at: Timestamp,
        proposed_change_id: str | None,
    ) -> tuple[Timestamp | None, Timestamp | None]:
        if from_time:
            from_timestamp = Timestamp(from_time.isoformat())
        # don't set default time if filtering on proposed change
        elif proposed_change_id:
            from_timestamp = None
        else:
            from_timestamp = branch_start_timestamp
        if to_time:
            to_timestamp = Timestamp(to_time.isoformat())
        # don't set default time if filtering on proposed change
        elif proposed_change_id:
            to_timestamp = None
        else:
            to_timestamp = at
        return from_timestamp, to_timestamp

    async def _get_branch_info(
        self,
        db: InfrahubDatabase,
        branch: str | None,
        proposed_change_id: str | None,
    ) -> DiffBranchInfo:
        try:
            diff_branch = await registry.get_branch(db=db, branch=branch)
            return DiffBranchInfo(
                name=diff_branch.name,
                branch_start_timestamp=Timestamp(diff_branch.get_branched_from()),
                is_terminal=diff_branch.is_terminal,
            )
        except BranchNotFoundError:
            if not branch and not proposed_change_id:
                raise ValidationError("Must include the branch or proposed_change_id argument") from None
            # case for deleted branch with an input proposed_change_id
            if branch and proposed_change_id:
                return DiffBranchInfo(name=branch, branch_start_timestamp=None, is_terminal=False)
            raise

    def _get_diff_query_params(
        self,
        branch_info: DiffBranchInfo,
        at: Timestamp,
        from_time: datetime | None = None,
        to_time: datetime | None = None,
        name: str | None = None,
        proposed_change_id: str | None = None,
    ) -> DiffQueryParams:
        # For terminal branches with no explicit time filters,
        # return the latest stored diff for the branch regardless of time
        if branch_info.is_terminal and not from_time and not to_time:
            return DiffQueryParams(
                from_time=None,
                to_time=None,
                tracking_id=BranchTrackingId(branch_info.name),
                exclude_merged=False,
            )

        from_timestamp, to_timestamp = self._get_timestamp(
            from_time=from_time,
            to_time=to_time,
            branch_start_timestamp=branch_info.branch_start_timestamp,
            at=at,
            proposed_change_id=proposed_change_id,
        )
        return DiffQueryParams(
            from_time=from_timestamp,
            to_time=to_timestamp,
            tracking_id=NameTrackingId(name) if name else None,
            exclude_merged=branch_info.is_terminal or not proposed_change_id,
        )

    async def resolve(
        self,
        root: dict,  # noqa: ARG002
        info: GraphQLResolveInfo,
        branch: str | None = None,
        name: str | None = None,
        from_time: datetime | None = None,
        to_time: datetime | None = None,
        filters: dict | None = None,
        include_parents: bool = True,
        root_node_uuids: list[str] | None = None,
        limit: int | None = None,
        offset: int | None = None,
        proposed_change_id: str | None = None,
    ) -> list[dict[str, Any]] | dict[str, Any] | None:
        component_registry = get_component_registry()
        graphql_context: GraphqlContext = info.context
        base_branch = await registry.get_branch(db=graphql_context.db, branch=registry.default_branch)

        branch_info = await self._get_branch_info(
            db=graphql_context.db, branch=branch, proposed_change_id=proposed_change_id
        )
        query_params = self._get_diff_query_params(
            branch_info=branch_info,
            at=graphql_context.at or Timestamp(),
            from_time=from_time,
            to_time=to_time,
            name=name,
            proposed_change_id=proposed_change_id,
        )

        diff_repo = await component_registry.get_component(DiffRepository, db=graphql_context.db, branch=base_branch)

        # Convert filters to dict and merge root_node_uuids for compatibility
        filters_dict = dict(filters or {})
        if root_node_uuids and "ids" in filters_dict and isinstance(filters_dict["ids"], list):
            filters_dict["ids"].extend(root_node_uuids)
        elif root_node_uuids:
            filters_dict["ids"] = root_node_uuids

        # ensure any ongoing diff updates complete before we try to retrieve them
        lock_request_time = Timestamp()
        diff_locker = DiffLocker()
        async with (
            diff_locker.acquire_lock(
                target_branch_name=base_branch.name, source_branch_name=branch_info.name, is_incremental=True
            ),
            diff_locker.acquire_lock(
                target_branch_name=base_branch.name, source_branch_name=branch_info.name, is_incremental=False
            ),
        ):
            lock_acquired_time = Timestamp()
            with trace.get_tracer(__name__).start_as_current_span("diff_tree_request") as span:
                get_start_time = Timestamp()
                enriched_diffs = await diff_repo.get(
                    base_branch_name=base_branch.name,
                    diff_branch_names=[branch_info.name],
                    from_time=query_params.from_time,
                    to_time=query_params.to_time,
                    filters=EnrichedDiffQueryFilters(**filters_dict),
                    include_parents=include_parents,
                    limit=limit,
                    offset=offset,
                    tracking_id=query_params.tracking_id,
                    include_empty=True,
                    proposed_change_id=proposed_change_id,
                    exclude_merged=query_params.exclude_merged,
                )
                get_end_time = Timestamp()

                span.set_attribute("base_branch_name", base_branch.name)
                span.set_attribute("diff_branch_name", branch_info.name)
                span.set_attribute(
                    "from_time", query_params.from_time.to_string() if query_params.from_time else "null"
                )
                span.set_attribute("to_time", query_params.to_time.to_string() if query_params.to_time else "null")
                span.set_attribute("proposed_change_id", proposed_change_id or "null")
                span.set_attribute("lock_request_time", lock_request_time.to_string())
                span.set_attribute("lock_acquired_time", lock_acquired_time.to_string())
                span.set_attribute("get_start_time", get_start_time.to_string())
                span.set_attribute("get_end_time", get_end_time.to_string())
                span.set_attribute("num_diffs", str(len(enriched_diffs)))

        if not enriched_diffs:
            return None
        if len(enriched_diffs) > 0:
            # take the one with the longest duration that covers multiple branches
            enriched_diff = sorted(
                enriched_diffs,
                key=lambda d: (
                    d.base_branch_name != d.diff_branch_name,
                    d.to_time.to_datetime() - d.from_time.to_datetime(),
                ),
                reverse=True,
            )[0]
        else:
            enriched_diff = enriched_diffs[0]

        full_fields = extract_graphql_fields(info=info)
        diff_tree = await self.to_diff_tree(enriched_diff_root=enriched_diff, graphql_context=graphql_context)
        need_base_changes = "num_untracked_base_changes" in full_fields
        need_branch_changes = "num_untracked_diff_changes" in full_fields
        if need_base_changes or need_branch_changes:
            await self._add_untracked_fields(
                db=graphql_context.db,
                diff_response=diff_tree,
                from_time=enriched_diff.to_time,
                base_branch_name=base_branch.name if need_base_changes else None,
                diff_branch_name=branch_info.name if need_branch_changes else None,
            )
        return await self.to_graphql(fields=full_fields, diff_object=diff_tree)

    async def summary(
        self,
        root: dict,  # noqa: ARG002
        info: GraphQLResolveInfo,
        branch: str | None = None,
        from_time: datetime | None = None,
        to_time: datetime | None = None,
        filters: dict | None = None,
        proposed_change_id: str | None = None,
    ) -> list[dict[str, Any]] | dict[str, Any] | None:
        component_registry = get_component_registry()
        graphql_context: GraphqlContext = info.context
        base_branch = await registry.get_branch(db=graphql_context.db, branch=registry.default_branch)

        branch_info = await self._get_branch_info(
            db=graphql_context.db, branch=branch, proposed_change_id=proposed_change_id
        )
        query_params = self._get_diff_query_params(
            branch_info=branch_info,
            at=graphql_context.at or Timestamp(),
            from_time=from_time,
            to_time=to_time,
            proposed_change_id=proposed_change_id,
        )

        diff_repo = await component_registry.get_component(DiffRepository, db=graphql_context.db, branch=base_branch)

        filters_dict = dict(filters or {})

        # ensure any ongoing diff updates complete before we try to retrieve them
        lock_request_time = Timestamp()
        diff_locker = DiffLocker()
        async with (
            diff_locker.acquire_lock(
                target_branch_name=base_branch.name, source_branch_name=branch_info.name, is_incremental=True
            ),
            diff_locker.acquire_lock(
                target_branch_name=base_branch.name, source_branch_name=branch_info.name, is_incremental=False
            ),
        ):
            lock_acquired_time = Timestamp()
            with trace.get_tracer(__name__).start_as_current_span("diff_tree_summary_request") as span:
                get_start_time = Timestamp()
                summary = await diff_repo.summary(
                    base_branch_name=base_branch.name,
                    diff_branch_names=[branch_info.name],
                    from_time=query_params.from_time,
                    to_time=query_params.to_time,
                    tracking_id=query_params.tracking_id,
                    filters=filters_dict,
                    proposed_change_id=proposed_change_id,
                    exclude_merged=query_params.exclude_merged,
                )
                get_end_time = Timestamp()

                span.set_attribute("base_branch_name", base_branch.name)
                span.set_attribute("diff_branch_name", branch_info.name)
                span.set_attribute(
                    "from_time", query_params.from_time.to_string() if query_params.from_time else "null"
                )
                span.set_attribute("to_time", query_params.to_time.to_string() if query_params.to_time else "null")
                span.set_attribute("proposed_change_id", proposed_change_id or "null")
                span.set_attribute("lock_request_time", lock_request_time.to_string())
                span.set_attribute("lock_acquired_time", lock_acquired_time.to_string())
                span.set_attribute("get_start_time", get_start_time.to_string())
                span.set_attribute("get_end_time", get_end_time.to_string())
                span.set_attribute("summary_retrieved", str(summary is not None))

        if summary is None:
            return None

        diff_tree_summary = DiffTreeSummary(
            base_branch=base_branch.name,
            diff_branch=branch_info.name,
            from_time=summary.from_time.to_datetime(),
            to_time=summary.to_time.to_datetime(),
            **summary.model_dump(exclude={"from_time", "to_time"}),
        )
        full_fields = extract_graphql_fields(info=info)
        need_base_changes = "num_untracked_base_changes" in full_fields
        need_branch_changes = "num_untracked_diff_changes" in full_fields
        if need_base_changes or need_branch_changes:
            await self._add_untracked_fields(
                db=graphql_context.db,
                diff_response=diff_tree_summary,
                from_time=summary.to_time,
                base_branch_name=base_branch.name if need_base_changes else None,
                diff_branch_name=branch_info.name if need_branch_changes else None,
            )
        return diff_tree_summary


class IncExclFilterOptions(InputObjectType):
    includes = List(String)
    excludes = List(String)


class IncExclFilterStatusOptions(InputObjectType):
    includes = List(GrapheneDiffActionEnum)
    excludes = List(GrapheneDiffActionEnum)


class DiffTreeQueryFilters(InputObjectType):
    ids = List(String)
    status = IncExclFilterStatusOptions()
    kind = IncExclFilterOptions()
    namespace = IncExclFilterOptions()


DiffTreeQuery = Field(
    DiffTree,
    name=String(),
    branch=String(),
    from_time=DateTime(),
    to_time=DateTime(),
    root_node_uuids=Argument(List(String), deprecation_reason="replaced by filters"),
    include_parents=Boolean(),
    filters=DiffTreeQueryFilters(),
    limit=Int(),
    offset=Int(),
    proposed_change_id=String(),
    resolver=DiffTreeResolver().resolve,
    required=False,
)

DiffTreeSummaryQuery = Field(
    DiffTreeSummary,
    name=String(),
    branch=String(),
    from_time=DateTime(),
    to_time=DateTime(),
    filters=DiffTreeQueryFilters(),
    proposed_change_id=String(),
    resolver=DiffTreeResolver().summary,
    required=False,
)



from __future__ import annotations

from typing import TYPE_CHECKING

from graphene import Boolean, InputField, InputObjectType, Mutation, String

from infrahub.core.constants import BranchConflictKeep, InfrahubKind
from infrahub.core.diff.model.path import ConflictSelection
from infrahub.core.diff.repository.repository import DiffRepository
from infrahub.core.manager import NodeManager
from infrahub.database import retry_db_transaction
from infrahub.dependencies.registry import get_component_registry
from infrahub.exceptions import ProcessingError
from infrahub.graphql.context import apply_external_context
from infrahub.graphql.enums import ConflictSelection as GraphQlConflictSelection
from infrahub.graphql.types.context import ContextInput

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from ..initialization import GraphqlContext


class ResolveDiffConflictInput(InputObjectType):
    conflict_id = InputField(String(required=True), description="ID of the diff conflict to resolve")
    selected_branch = InputField(
        GraphQlConflictSelection(required=True), description="Which version of the conflict to select"
    )


class ResolveDiffConflict(Mutation):
    class Arguments:
        data = ResolveDiffConflictInput(required=True)
        context = ContextInput(required=False)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="resolve_diff_conflict")
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: ResolveDiffConflictInput,
        context: ContextInput | None = None,
    ) -> ResolveDiffConflict:
        graphql_context: GraphqlContext = info.context
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        component_registry = get_component_registry()
        diff_repo = await component_registry.get_component(
            DiffRepository, db=graphql_context.db, branch=graphql_context.branch
        )

        selection = ConflictSelection(data.selected_branch.value) if data.selected_branch else None
        conflict = await diff_repo.get_conflict_by_id(conflict_id=data.conflict_id)
        if conflict.resolvable is False:
            raise ProcessingError("Conflict must be resolved manually on conflicting branch(es)")
        await diff_repo.update_conflict_by_id(conflict_id=data.conflict_id, selection=selection)

        core_data_checks = await NodeManager.query(
            db=graphql_context.db,
            schema=InfrahubKind.DATACHECK,
            filters={"enriched_conflict_id__value": data.conflict_id},
        )
        if not core_data_checks:
            return cls(ok=True)
        if data.selected_branch is GraphQlConflictSelection.BASE_BRANCH:
            keep_branch = BranchConflictKeep.TARGET
        elif data.selected_branch is GraphQlConflictSelection.DIFF_BRANCH:
            keep_branch = BranchConflictKeep.SOURCE
        else:
            keep_branch = None
        for cdc in core_data_checks:
            cdc.keep_branch.value = keep_branch
            await cdc.save(db=graphql_context.db)
        return cls(ok=True)



from infrahub.core.diff.enricher.cardinality_one import DiffCardinalityOneEnricher
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class DiffCardinalityOneEnricherDependency(DependencyBuilder[DiffCardinalityOneEnricher]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> DiffCardinalityOneEnricher:
        return DiffCardinalityOneEnricher(db=context.db)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.query import Query, QueryType

from ..shared import GraphMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class RemoveGenericGenerateTemplateQuery(Query):
    """Remove generate_template Attribute nodes from all SchemaGeneric-labeled instances."""

    name = "migration_065_remove_generic_generate_template"
    type: QueryType = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
MATCH (sg:SchemaGeneric)-[:HAS_ATTRIBUTE]->(attr:Attribute {name: "generate_template"})
WITH DISTINCT attr
DETACH DELETE attr
        """
        self.add_to_query(query)


class RemoveGenericGenerateTemplateSchemaAttributeQuery(Query):
    """Remove the generate_template SchemaAttribute from the SchemaGeneric type definition."""

    name = "migration_065_remove_generic_generate_template_schema_attribute"
    type: QueryType = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
// Find the SchemaGeneric type definition (stored as SchemaNode)
MATCH p1 = (sn:SchemaNode)-[:HAS_ATTRIBUTE]->(:Attribute {name: "name"})-[:HAS_VALUE]->(:AttributeValueIndexed {value: "Generic"})
WHERE all(r IN relationships(p1) WHERE r.status = "active" AND r.to IS NULL)
MATCH p2 = (sn)-[:HAS_ATTRIBUTE]->(:Attribute {name: "namespace"})-[:HAS_VALUE]->(:AttributeValueIndexed {value: "Schema"})
WHERE all(r IN relationships(p2) WHERE r.status = "active" AND r.to IS NULL)
WITH sn
LIMIT 1
// Find the generate_template SchemaAttribute child
MATCH p3 = (sn)-[:IS_RELATED]-(rel:Relationship {name: "schema__node__attributes"})
               -[:IS_RELATED]-(sa:SchemaAttribute)
               -[:HAS_ATTRIBUTE]->(:Attribute {name: "name"})
               -[:HAS_VALUE]->(:AttributeValueIndexed {value: "generate_template"})
WHERE all(r IN relationships(p3) WHERE r.status = "active" AND r.to IS NULL)
WITH sa, rel
LIMIT 1
// Find child Attribute nodes and their value nodes
MATCH (sa)-[:HAS_ATTRIBUTE]->(attr:Attribute)-[:HAS_VALUE]->(val)
// Delete the SchemaAttribute, Relationship node, and child Attributes
DETACH DELETE sa, rel, attr
// Clean up orphaned value nodes (Boolean nodes are shared and will still have edges)
WITH val
WHERE NOT EXISTS { MATCH (val)-[]-() }
DELETE val
        """
        self.add_to_query(query)


class Migration065(GraphMigration):
    name: str = "065_remove_generic_generate_template"
    minimum_version: int = 64
    queries: Sequence[type[Query]] = [
        RemoveGenericGenerateTemplateQuery,
        RemoveGenericGenerateTemplateSchemaAttributeQuery,
    ]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from infrahub.core.constants import SYSTEM_USER_ID, NumberPoolType
from infrahub.core.manager import NodeManager
from infrahub.core.protocols import CoreNumberPool
from infrahub.core.registry import registry
from infrahub.core.schema.attribute_parameters import NumberPoolParameters
from infrahub.log import get_logger
from infrahub.pools.registration import get_branches_with_schema_number_pool

if TYPE_CHECKING:
    from logging import Logger, LoggerAdapter

    from structlog.stdlib import BoundLogger

    from infrahub.core.schema import GenericSchema, NodeSchema
    from infrahub.core.schema.manager import SchemaManager
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.database import InfrahubDatabase
    from infrahub.pools.schema_number_pool_upserter import SchemaNumberPoolUpserter

default_log = get_logger()


class SchemaNumberPoolSynchronizer:
    """Synchronizes schema-defined number pools.

    This class handles the lifecycle of CoreNumberPool instances that are defined
    via NumberPool attributes in the schema. It ensures pools are created, updated,
    or deleted as the schema changes across branches.

    Args:
        db: Database connection.
        log: Logger instance.
        schema_manager: Schema manager for looking up schemas.
    """

    def __init__(
        self,
        db: InfrahubDatabase,
        schema_manager: SchemaManager,
        upserter: SchemaNumberPoolUpserter,
        log: Logger | LoggerAdapter | BoundLogger | None = None,
    ) -> None:
        self.db = db
        self.log = log or default_log
        self.schema_manager = schema_manager
        self.upserter = upserter

    async def run(self, user_id: str = SYSTEM_USER_ID) -> set[str]:
        """Execute the full synchronization process.

        Returns:
            Set of branch names where schema changes were persisted.
        """
        await self._sync_existing_pools_with_schema(user_id=user_id)
        return await self._process_all_branches(user_id=user_id)

    async def _sync_existing_pools_with_schema(self, user_id: str) -> None:
        """Update or delete existing pools based on current schema definitions."""
        schema_number_pools = await NodeManager.query(
            db=self.db,
            schema=CoreNumberPool,
            filters={"pool_type__value": NumberPoolType.SCHEMA.value},
            branch_agnostic=True,
        )
        for schema_number_pool in schema_number_pools:
            defined_on_branches = get_branches_with_schema_number_pool(
                kind=schema_number_pool.node.value, attribute_name=schema_number_pool.node_attribute.value
            )
            if registry.default_branch in defined_on_branches:
                await self._update_pool_from_schema(schema_number_pool, user_id=user_id)
            elif not defined_on_branches:
                self.log.info(
                    f"Deleting number pool (id={schema_number_pool.id}) as it is no longer defined in the schema"
                )
                await schema_number_pool.delete(db=self.db, user_id=user_id)

    async def _update_pool_from_schema(self, schema_number_pool: CoreNumberPool, user_id: str = SYSTEM_USER_ID) -> None:
        """Update a pool's range parameters if they differ from the schema."""
        schema = self.schema_manager.get(
            name=schema_number_pool.node.value, branch=registry.default_branch, duplicate=False
        )
        attribute = schema.get_attribute(name=schema_number_pool.node_attribute.value)
        number_pool_updated = False

        if isinstance(attribute.parameters, NumberPoolParameters):
            if schema_number_pool.start_range.value != attribute.parameters.start_range:
                schema_number_pool.start_range.value = attribute.parameters.start_range
                number_pool_updated = True
            if schema_number_pool.end_range.value != attribute.parameters.end_range:
                schema_number_pool.end_range.value = attribute.parameters.end_range
                number_pool_updated = True

        if number_pool_updated:
            self.log.info(
                f"Updating NumberPool={schema_number_pool.id} based on changes in the schema on {registry.default_branch}"
            )
            await schema_number_pool.save(db=self.db, user_id=user_id)

    async def _process_all_branches(self, user_id: str) -> set[str]:
        """Process all branches to create any missing number pools.

        Returns:
            Set of branch names where schema changes were persisted.
        """
        updated_branches: set[str] = set()

        for branch_name in self.schema_manager.get_branches():
            schemas_to_update: list[str] = []
            schema_branch = self.schema_manager.get_schema_branch(name=branch_name)

            # Process generics first so their pool IDs are available for inheriting nodes
            for generic_name in schema_branch.generic_names:
                generic_schema = schema_branch.get_generic(name=generic_name, duplicate=False)
                updated_schema = await self._process_schema_node(
                    schema_node=generic_schema,
                    branch_name=branch_name,
                    schema_branch=schema_branch,
                    user_id=user_id,
                )
                if updated_schema:
                    schema_branch.set(name=generic_schema.kind, schema=updated_schema)
                    schemas_to_update.append(generic_schema.kind)

            # Process nodes after generics - inherited attributes will look up pool IDs from generics
            for node_name in schema_branch.node_names:
                node_schema = schema_branch.get_node(name=node_name, duplicate=False)
                updated_schema = await self._process_schema_node(
                    schema_node=node_schema,
                    branch_name=branch_name,
                    schema_branch=schema_branch,
                    user_id=user_id,
                )
                if updated_schema:
                    schema_branch.set(name=node_schema.kind, schema=updated_schema)
                    schemas_to_update.append(node_schema.kind)

            if schemas_to_update:
                self.log.info(f"Persisting schema changes to the database on {branch_name}")
                await self.schema_manager.update_schema_branch(
                    db=self.db, branch=branch_name, schema=schema_branch, limit=schemas_to_update, update_db=True
                )
                updated_branches.add(branch_name)

        return updated_branches

    async def _process_schema_node(
        self,
        schema_node: NodeSchema | GenericSchema,
        branch_name: str,
        schema_branch: SchemaBranch,
        user_id: str,
    ) -> NodeSchema | GenericSchema | None:
        """Process NumberPool attributes for a schema node, creating pools as needed.

        Args:
            schema_node: The schema node to process.
            branch_name: The branch name for schema lookups.
            schema_branch: The SchemaBranch for schema lookups.
            user_id: The user ID for any save operations.

        Returns a NodeSchema or GenericSchema if the schema was updated.
        """
        updated_schema: NodeSchema | GenericSchema | None = None

        for attribute_name in schema_node.attribute_names:
            attribute = schema_node.get_attribute(name=attribute_name)
            if not isinstance(attribute.parameters, NumberPoolParameters):
                continue

            if attribute.parameters.number_pool_id:
                # Pool ID already exists, so the pool exists, move on
                continue

            # Try to get existing pool ID
            new_pool_id = await self.upserter.get_existing_number_pool_id(
                schema_node=schema_node,
                attribute=attribute,
                branch_name=branch_name,
                schema_branch=schema_branch,
            )

            # If no pool ID, create one
            if not new_pool_id:
                new_pool = await self.upserter.upsert_number_pool(
                    schema_node=schema_node,
                    attribute=attribute,
                    branch_name=branch_name,
                    schema_branch=schema_branch,
                    user_id=user_id,
                )
                new_pool_id = new_pool.id

            self.log.info(f"Setting {schema_node.kind}.{attribute_name} number_pool_id to {new_pool_id}")
            if not updated_schema:
                updated_schema = schema_node.duplicate()
            updated_attribute = updated_schema.get_attribute(name=attribute_name)
            attribute_parameters = cast("NumberPoolParameters", updated_attribute.parameters)
            attribute_parameters.number_pool_id = new_pool_id

        return updated_schema



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core.constants import GlobalPermissions, InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.exceptions import NodeNotFoundError, ValidationError
from infrahub.permissions import define_global_permission_from_branch

if TYPE_CHECKING:
    from .initialization import GraphqlContext
    from .types.context import ContextInput


async def apply_external_context(graphql_context: GraphqlContext, context_input: ContextInput | None) -> None:
    """Applies context provided by an external mutation to the GraphQL context"""
    if not context_input or not context_input.account or not context_input.account.id:
        return

    if graphql_context.active_account_session.account_id == context_input.account.id:
        # If the account_id from the request context is the same as the current account
        # there's no point moving forward with other checks to override the current
        # context we can just continue with what is already there.
        return

    permission = define_global_permission_from_branch(
        permission=GlobalPermissions.OVERRIDE_CONTEXT, branch_name=graphql_context.branch.name
    )

    graphql_context.active_permissions.raise_for_permission(permission=permission)

    try:
        account = await NodeManager.get_one_by_id_or_default_filter(
            db=graphql_context.db, id=str(context_input.account.id), kind=InfrahubKind.GENERICACCOUNT
        )
    except NodeNotFoundError as exc:
        raise ValidationError(input_value="Unable to set context for account that doesn't exist") from exc

    graphql_context.active_account_session.account_id = account.id



from datetime import datetime
from pathlib import Path
from typing import BinaryIO

VARIABLE_TYPE_MAPPING = (
    (str, "String!"),
    (str | None, "String"),
    (int, "Int!"),
    (int | None, "Int"),
    (float, "Float!"),
    (float | None, "Float"),
    (bool, "Boolean!"),
    (bool | None, "Boolean"),
    (datetime, "DateTime!"),
    (datetime | None, "DateTime"),
    (bytes, "Upload!"),
    (Path, "Upload!"),
    (BinaryIO, "Upload!"),
)



from infrahub.core.relationship.constraints.peer_parent import RelationshipPeerParentConstraint
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class RelationshipPeerParentConstraintDependency(DependencyBuilder[RelationshipPeerParentConstraint]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> RelationshipPeerParentConstraint:
        return RelationshipPeerParentConstraint(db=context.db, branch=context.branch)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.query import Query, QueryType

from ..shared import GraphMigration, MigrationInput

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class RemoveIsVisibleRelationshipQuery(Query):
    name = "remove_is_visible_relationship"
    type: QueryType = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
        MATCH ()-[rel:IS_VISIBLE]->()
        CALL (rel) {
          DELETE rel
        } IN TRANSACTIONS
        """
        self.add_to_query(query)


class RemoveIsVisibleFromDiffsQuery(Query):
    name = "remove_is_visible_from_diffs"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        query = """
        MATCH (diff_prop:DiffProperty {property_type: "IS_VISIBLE"})
        CALL (diff_prop) {
            DETACH DELETE diff_prop
        } IN TRANSACTIONS
        """
        self.add_to_query(query)


class Migration049(GraphMigration):
    name: str = "049_remove_is_visible_relationship"
    minimum_version: int = 48
    queries: Sequence[type[Query]] = [RemoveIsVisibleRelationshipQuery, RemoveIsVisibleFromDiffsQuery]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        return await self.do_execute(migration_input=migration_input)



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core.constants.schema import RESOURCE_POOL_REL_SUFFIX
from infrahub.exceptions import ValidationError

from .interface import RelationshipManagerConstraintInterface

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.relationship.model import RelationshipManager
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.database import InfrahubDatabase


class TemplateResourcePoolExclusiveConstraint(RelationshipManagerConstraintInterface):
    """Constraint that prevents setting both a relationship/attribute and its _from_resource_pool counterpart on templates.

    On template instances, users can either:
    - Set a fixed value (relationship peer or attribute value)
    - Set a pool to allocate from (e.g., ip_address_from_resource_pool or weight_from_resource_pool)

    But not both at the same time.
    """

    def __init__(self, db: InfrahubDatabase, branch: Branch | None = None) -> None:
        self.db = db
        self.branch = branch

    def expand_filters(self, field_filters: list[str], node_schema: MainSchemaTypes) -> list[str]:
        if not node_schema.is_template_schema:
            return field_filters

        expanded = list(field_filters)
        for name in field_filters:
            if name not in node_schema.attribute_names:
                continue
            pool_rel_name = f"{name}{RESOURCE_POOL_REL_SUFFIX}"
            if pool_rel_name in node_schema.relationship_names and pool_rel_name not in expanded:
                expanded.append(pool_rel_name)
        return expanded

    async def check(self, relm: RelationshipManager, node_schema: MainSchemaTypes, node: Node) -> None:
        if not node_schema.is_template_schema:
            return

        if not await self._relationship_has_peers(relm=relm):
            return

        rel_name = relm.schema.name

        if rel_name.endswith(RESOURCE_POOL_REL_SUFFIX):
            original_name = rel_name.removesuffix(RESOURCE_POOL_REL_SUFFIX)
            if original_name in node_schema.relationship_names:
                await self._check_counterpart_not_set(node=node, counterpart_name=original_name, current_name=rel_name)
            elif original_name in node_schema.attribute_names:
                self._check_attribute_counterpart_not_set(
                    node=node, attribute_name=original_name, current_name=rel_name
                )
        else:
            # Check if this relationship has a pool counterpart and if it's set
            pool_rel_name = f"{rel_name}{RESOURCE_POOL_REL_SUFFIX}"
            if pool_rel_name in node_schema.relationship_names:
                await self._check_counterpart_not_set(node=node, counterpart_name=pool_rel_name, current_name=rel_name)

    async def _relationship_has_peers(self, relm: RelationshipManager) -> bool:
        """Check if a relationship currently has peers (locally or untouched in DB)."""
        update_details = await relm.fetch_relationship_ids(db=self.db, force_refresh=True)
        peers_in_local_state = update_details.peer_ids_present_both or update_details.peer_ids_present_local_only
        peers_untouched_in_db = update_details.peer_ids_present_database_only and not relm.has_fetched_relationships
        return bool(peers_in_local_state or peers_untouched_in_db)

    async def _check_counterpart_not_set(self, node: Node, counterpart_name: str, current_name: str) -> None:
        """Check that the counterpart relationship is not set."""
        try:
            counterpart_relm = node.get_relationship(name=counterpart_name)
        except ValueError:
            return

        if await self._relationship_has_peers(relm=counterpart_relm):
            raise ValidationError(
                {
                    current_name: (
                        f"Cannot set '{current_name}' when '{counterpart_name}' is already set. "
                        "Templates can only use one of: direct relationship or resource pool allocation."
                    )
                }
            )

    def _check_attribute_counterpart_not_set(self, node: Node, attribute_name: str, current_name: str) -> None:
        """Check that the counterpart attribute does not have a user-set value."""
        try:
            attr = node.get_attribute(name=attribute_name)
        except ValueError:
            return

        if not attr.is_default and attr.value is not None:
            raise ValidationError(
                {
                    current_name: (
                        f"Cannot set '{current_name}' when '{attribute_name}' has a value set. "
                        "Templates can only use one of: direct attribute value or resource pool allocation."
                    )
                }
            )



from infrahub.core.constants import (
    BranchSupportType,
    InfrahubKind,
)
from infrahub.core.constants import RelationshipCardinality as Cardinality
from infrahub.core.constants import RelationshipKind as RelKind

from ...attribute_schema import AttributeSchema as Attr
from ...node_schema import NodeSchema
from ...relationship_schema import (
    RelationshipSchema as Rel,
)

core_check_definition = NodeSchema(
    name="CheckDefinition",
    namespace="Core",
    description="Defines a user-defined check that validates data in a proposed change",
    include_in_menu=False,
    icon="mdi:check-all",
    label="Check Definition",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    branch=BranchSupportType.AWARE,
    uniqueness_constraints=[["name__value"]],
    generate_profile=False,
    inherit_from=[InfrahubKind.TASKTARGET],
    attributes=[
        Attr(name="name", kind="Text", unique=True),
        Attr(name="description", kind="Text", optional=True),
        Attr(name="file_path", kind="Text", description="Path to the Python file containing the check class"),
        Attr(name="class_name", kind="Text", description="Name of the Python class implementing the check"),
        Attr(
            name="timeout",
            kind="Number",
            description="Maximum execution time in seconds before the check times out",
            default_value=60,
        ),
        Attr(name="parameters", kind="JSON", description="Additional parameters to pass to the check", optional=True),
    ],
    relationships=[
        Rel(
            name="repository",
            peer=InfrahubKind.GENERICREPOSITORY,
            kind=RelKind.ATTRIBUTE,
            cardinality=Cardinality.ONE,
            identifier="check_definition__repository",
            optional=False,
        ),
        Rel(
            name="query",
            peer=InfrahubKind.GRAPHQLQUERY,
            kind=RelKind.ATTRIBUTE,
            identifier="check_definition__graphql_query",
            cardinality=Cardinality.ONE,
            optional=True,
        ),
        Rel(
            name="targets",
            peer=InfrahubKind.GENERICGROUP,
            kind=RelKind.ATTRIBUTE,
            identifier="check_definition___group",
            cardinality=Cardinality.ONE,
            optional=True,
        ),
        Rel(
            name="tags",
            peer=InfrahubKind.TAG,
            kind=RelKind.ATTRIBUTE,
            optional=True,
            cardinality=Cardinality.MANY,
        ),
    ],
)



import ipaddress
from enum import Enum

IPNetworkType = ipaddress.IPv6Network | ipaddress.IPv4Network
IPAddressType = ipaddress.IPv6Interface | ipaddress.IPv4Interface
AllIPTypes = IPNetworkType | IPAddressType


class PrefixMemberType(Enum):
    PREFIX = "prefix"
    ADDRESS = "address"



from infrahub.core.constants import (
    BranchSupportType,
    InfrahubKind,
    RelationshipDeleteBehavior,
)
from infrahub.core.constants import RelationshipCardinality as Cardinality
from infrahub.core.constants import RelationshipKind as RelKind
from infrahub.proposed_change.constants import ProposedChangeState

from ...attribute_schema import AttributeSchema as Attr
from ...node_schema import NodeSchema
from ...relationship_schema import (
    RelationshipSchema as Rel,
)

core_proposed_change = NodeSchema(
    name="ProposedChange",
    namespace="Core",
    description="Metadata related to a proposed change",
    include_in_menu=False,
    icon="mdi:file-replace-outline",
    label="Proposed Change",
    default_filter="name__value",
    display_labels=["name__value"],
    generate_profile=False,
    branch=BranchSupportType.AGNOSTIC,
    inherit_from=[InfrahubKind.TASKTARGET],
    documentation="/topics/proposed-change",
    attributes=[
        Attr(name="name", kind="Text", optional=False),
        Attr(name="description", kind="TextArea", optional=True),
        Attr(
            name="source_branch", kind="Text", description="Name of the branch containing the changes", optional=False
        ),
        Attr(
            name="destination_branch",
            kind="Text",
            description="Name of the branch to merge changes into",
            optional=False,
        ),
        Attr(
            name="state",
            kind="Text",
            description="Current state of the proposed change",
            enum=ProposedChangeState.available_types(),
            default_value=ProposedChangeState.OPEN.value,
            optional=True,
        ),
        Attr(
            name="is_draft",
            kind="Boolean",
            description="Whether the proposed change is still a draft",
            optional=False,
            default_value=False,
        ),
        # Ideally we should support some "runtime-attribute" that could not even be stored in the database.
        Attr(
            name="total_comments",
            kind="Number",
            description="Total number of comments on this proposed change",
            optional=True,
            read_only=True,
        ),
    ],
    relationships=[
        Rel(
            name="approved_by",
            peer=InfrahubKind.GENERICACCOUNT,
            optional=True,
            cardinality=Cardinality.MANY,
            kind=RelKind.ATTRIBUTE,
            branch=BranchSupportType.AGNOSTIC,
            identifier="coreaccount__proposedchange_approved_by",
            read_only=True,
        ),
        Rel(
            name="rejected_by",
            peer=InfrahubKind.GENERICACCOUNT,
            optional=True,
            cardinality=Cardinality.MANY,
            kind=RelKind.ATTRIBUTE,
            branch=BranchSupportType.AGNOSTIC,
            identifier="coreaccount__proposedchange_rejected_by",
            read_only=True,
        ),
        Rel(
            name="reviewers",
            peer=InfrahubKind.GENERICACCOUNT,
            optional=True,
            cardinality=Cardinality.MANY,
            kind=RelKind.ATTRIBUTE,
            branch=BranchSupportType.AGNOSTIC,
            identifier="coreaccount__proposedchange_reviewed_by",
        ),
        Rel(
            name="comments",
            peer=InfrahubKind.CHANGECOMMENT,
            kind=RelKind.COMPONENT,
            optional=True,
            cardinality=Cardinality.MANY,
            on_delete=RelationshipDeleteBehavior.CASCADE,
        ),
        Rel(
            name="threads",
            peer=InfrahubKind.THREAD,
            identifier="proposedchange__thread",
            kind=RelKind.COMPONENT,
            optional=True,
            cardinality=Cardinality.MANY,
            on_delete=RelationshipDeleteBehavior.CASCADE,
        ),
        Rel(
            name="validations",
            peer=InfrahubKind.VALIDATOR,
            kind=RelKind.COMPONENT,
            identifier="proposed_change__validator",
            optional=True,
            cardinality=Cardinality.MANY,
            on_delete=RelationshipDeleteBehavior.CASCADE,
        ),
    ],
)



from __future__ import annotations

import hashlib
import io
from typing import TYPE_CHECKING

from fastapi import APIRouter, Depends, File, HTTPException, Request, Response, UploadFile
from infrahub_sdk.uuidt import UUIDT
from pydantic import BaseModel

from infrahub.api.dependencies import get_current_user, get_db
from infrahub.api.storage import file_object
from infrahub.core import registry
from infrahub.core.protocols import CoreFileObject
from infrahub.database import InfrahubDatabase  # noqa: TC001

if TYPE_CHECKING:
    from infrahub.auth import AccountSession


router = APIRouter(prefix="/storage")
router.include_router(file_object.router)


class UploadResponse(BaseModel):
    identifier: str
    checksum: str


class UploadContentPayload(BaseModel):
    content: str


@router.get("/object/{identifier:str}")
async def get_file(
    identifier: str,
    request: Request,
    db: InfrahubDatabase = Depends(get_db),
    _: AccountSession = Depends(get_current_user),
) -> Response:
    if await registry.manager.query(db=db, schema=CoreFileObject, filters={"storage_id__value": identifier}, limit=1):
        file_url = request.url_for("download_file_object_by_storage_id", storage_id=identifier)
        raise HTTPException(status_code=403, detail=f"Use {file_url.path} instead.")

    content = registry.storage.retrieve(identifier=identifier)
    return Response(content=content)


@router.post("/upload/content")
def upload_content(item: UploadContentPayload, _: str = Depends(get_current_user)) -> UploadResponse:
    file_content = bytes(item.content, encoding="utf-8")
    identifier = str(UUIDT())

    checksum = hashlib.md5(file_content, usedforsecurity=False).hexdigest()
    registry.storage.store(identifier=identifier, content=io.BytesIO(file_content))
    return UploadResponse(identifier=identifier, checksum=checksum)


@router.post("/upload/file")
def upload_file(file: UploadFile = File(...), _: AccountSession = Depends(get_current_user)) -> UploadResponse:
    identifier = str(UUIDT())

    hasher = hashlib.md5(usedforsecurity=False)
    while chunk := file.file.read(65536):
        hasher.update(chunk)
    checksum = hasher.hexdigest()

    file.file.seek(0)
    registry.storage.store(identifier=identifier, content=file.file)
    return UploadResponse(identifier=identifier, checksum=checksum)



from __future__ import annotations

from prefect import flow
from prefect.logging import get_run_logger

from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.pools.tasks import validate_schema_number_pools
from infrahub.services import InfrahubServices  # noqa: TC001  needed for prefect flow
from infrahub.workflows.utils import wait_for_schema_to_converge


@flow(
    name="schema-updated",
    flow_run_name="Running actions after the schema was updated on '{branch_name}'",
)
async def schema_updated(
    branch_name: str,
    schema_hash: str,  # noqa: ARG001
    context: InfrahubContext,
    service: InfrahubServices,
) -> None:
    log = get_run_logger()
    await wait_for_schema_to_converge(
        branch_name=branch_name, component=service.component, db=service.database, log=log
    )

    updated_branches = await validate_schema_number_pools(branch_name=branch_name, context=context, service=service)

    if updated_branches:
        await wait_for_schema_to_converge(
            branch_name=branch_name, component=service.component, db=service.database, log=log
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from fastapi import APIRouter, Depends, Path, Request
from starlette.responses import JSONResponse, PlainTextResponse

from infrahub.api.dependencies import (
    BranchParams,
    get_branch_params,
    get_context,
    get_current_user,
    get_db,
)
from infrahub.core.manager import NodeManager
from infrahub.core.protocols import (
    CoreGenericRepository,
    CoreGraphQLQuery,
    CoreTransformJinja2,
    CoreTransformPython,
)
from infrahub.database import InfrahubDatabase  # noqa: TC001
from infrahub.exceptions import TransformError
from infrahub.graphql.execution import execute_graphql_query
from infrahub.graphql.initialization import prepare_graphql_params
from infrahub.graphql.utils import extract_data
from infrahub.transformations.models import TransformJinjaTemplateData, TransformPythonData
from infrahub.workflows.catalogue import TRANSFORM_JINJA2_RENDER, TRANSFORM_PYTHON_RENDER

if TYPE_CHECKING:
    from infrahub.auth import AccountSession
    from infrahub.context import InfrahubContext
    from infrahub.services import InfrahubServices
router = APIRouter()


@router.get("/transform/python/{transform_id:str}")
async def transform_python(
    request: Request,
    transform_id: str,
    db: InfrahubDatabase = Depends(get_db),
    context: InfrahubContext = Depends(get_context),
    branch_params: BranchParams = Depends(get_branch_params),
    _: AccountSession = Depends(get_current_user),
) -> JSONResponse:
    params = {key: value for key, value in request.query_params.items() if key not in ["branch", "at"]}

    transform = await NodeManager.get_one_by_id_or_default_filter(
        db=db,
        id=transform_id,
        kind=CoreTransformPython,
        branch=branch_params.branch,
        at=branch_params.at,
    )

    query = await transform.query.get_peer(db=db, peer_type=CoreGraphQLQuery, raise_on_error=True)
    repository = await transform.repository.get_peer(db=db, peer_type=CoreGenericRepository, raise_on_error=True)

    if repository.commit.value is None:  # type: ignore[attr-defined]
        raise TransformError(
            repository_name=repository.name.value,
            location=repository.location.value,
            commit="n/a",
            message="Repository doesn't have a commit",
        )

    async with db.start_session(read_only=True) as dbs:
        gql_params = await prepare_graphql_params(
            db=dbs, branch=branch_params.branch, at=branch_params.at, service=request.app.state.service
        )

        result = await execute_graphql_query(
            schema=gql_params.schema,
            source=query.query.value,
            context_value=gql_params.context,
            root_value=None,
            variable_values=params,
        )

    data = extract_data(query_name=query.name.value, result=result)

    service: InfrahubServices = request.app.state.service

    message = TransformPythonData(
        repository_id=repository.id,
        repository_name=repository.name.value,
        repository_kind=repository.get_kind(),
        commit=repository.commit.value,  # type: ignore[attr-defined]
        branch=branch_params.branch.name,
        transform_location=f"{transform.file_path.value}::{transform.class_name.value}",
        timeout=transform.timeout.value,
        convert_query_response=transform.convert_query_response.value or False,
        data=data,
    )

    response = await service.workflow.execute_workflow(
        workflow=TRANSFORM_PYTHON_RENDER, context=context, parameters={"message": message}
    )
    return JSONResponse(content=response)


@router.get("/transform/jinja2/{transform_id}", response_class=PlainTextResponse)
async def transform_jinja2(
    request: Request,
    transform_id: str = Path(description="ID or Name of the Jinja2 Transform to render"),
    db: InfrahubDatabase = Depends(get_db),
    branch_params: BranchParams = Depends(get_branch_params),
    context: InfrahubContext = Depends(get_context),
    _: AccountSession = Depends(get_current_user),
) -> PlainTextResponse:
    params = {key: value for key, value in request.query_params.items() if key not in ["branch", "at"]}

    transform = await NodeManager.get_one_by_id_or_default_filter(
        db=db,
        id=transform_id,
        kind=CoreTransformJinja2,
        branch=branch_params.branch,
        at=branch_params.at,
    )

    query = await transform.query.get_peer(db=db, peer_type=CoreGraphQLQuery, raise_on_error=True)
    repository = await transform.repository.get_peer(db=db, peer_type=CoreGenericRepository, raise_on_error=True)

    if repository.commit.value is None:  # type: ignore[attr-defined]
        raise TransformError(
            repository_name=repository.name.value,
            location=repository.location.value,
            commit="n/a",
            message="Repository doesn't have a commit",
        )

    async with db.start_session(read_only=True) as dbs:
        gql_params = await prepare_graphql_params(
            db=dbs, branch=branch_params.branch, at=branch_params.at, service=request.app.state.service
        )

        result = await execute_graphql_query(
            schema=gql_params.schema,
            source=query.query.value,
            context_value=gql_params.context,
            root_value=None,
            variable_values=params,
        )

    data = extract_data(query_name=query.name.value, result=result)

    message = TransformJinjaTemplateData(
        repository_id=repository.id,
        repository_name=repository.name.value,
        repository_kind=repository.get_kind(),
        commit=repository.commit.value,  # type: ignore[attr-defined]
        branch=branch_params.branch.name,
        template_location=transform.template_path.value,
        timeout=transform.timeout.value,
        data=data,
    )

    service: InfrahubServices = request.app.state.service

    response = await service.workflow.execute_workflow(
        workflow=TRANSFORM_JINJA2_RENDER, context=context, expected_return=str, parameters={"message": message}
    )
    return PlainTextResponse(content=response)



from __future__ import annotations

import uuid
from collections import defaultdict
from typing import TYPE_CHECKING, Any
from uuid import UUID

from prefect.client.schemas.objects import Log as PrefectLog  # noqa: TC002
from prefect.events.filters import (
    EventFilter,
    EventIDFilter,
    EventNameFilter,
    EventOccurredFilter,
    EventRelatedFilter,
    EventResourceFilter,
)
from prefect.events.filters import EventOrder as PrefectEventOrder
from prefect.events.schemas.events import ResourceSpecification
from pydantic import BaseModel, Field

from infrahub.core.timestamp import Timestamp
from infrahub.events.constants import EventSortOrder

from .constants import LOG_LEVEL_MAPPING

if TYPE_CHECKING:
    from datetime import datetime


class RelatedNodeInfo(BaseModel):
    id: str
    kind: str | None = None


class RelatedNodesInfo(BaseModel):
    flows: dict[UUID, dict[str, RelatedNodeInfo]] = Field(default_factory=lambda: defaultdict(dict))  # type: ignore[arg-type]
    nodes: dict[str, RelatedNodeInfo] = Field(default_factory=dict)

    def add_nodes(self, flow_id: UUID, node_ids: list[str]) -> None:
        for node_id in node_ids:
            self.add_node(flow_id=flow_id, node_id=node_id)

    def add_node(self, flow_id: UUID, node_id: str) -> None:
        if node_id not in self.nodes:
            node = RelatedNodeInfo(id=node_id)
            self.nodes[node_id] = node
        self.flows[flow_id][node_id] = self.nodes[node_id]

    def get_related_nodes(self, flow_id: UUID) -> list[RelatedNodeInfo]:
        if flow_id not in self.flows or len(self.flows[flow_id].keys()) == 0:
            return []
        return list(self.flows[flow_id].values())

    def get_related_nodes_as_dict(self, flow_id: UUID) -> list[dict[str, str | None]]:
        if flow_id not in self.flows or len(self.flows[flow_id].keys()) == 0:
            return []
        return [item.model_dump() for item in list(self.flows[flow_id].values())]

    def get_first_related_node(self, flow_id: UUID) -> RelatedNodeInfo | None:
        if nodes := self.get_related_nodes(flow_id=flow_id):
            return nodes[0]
        return None

    def get_unique_related_node_ids(self) -> list[str]:
        return list(self.nodes.keys())


class FlowLogs(BaseModel):
    logs: defaultdict[UUID, list[PrefectLog]] = Field(default_factory=lambda: defaultdict(list))  # type: ignore[arg-type]

    def to_graphql(self, flow_id: UUID) -> list[dict]:
        return [
            {
                "node": {
                    "message": log.message,
                    "severity": LOG_LEVEL_MAPPING.get(log.level, "error"),
                    "timestamp": log.timestamp.isoformat(),
                }
            }
            for log in self.logs[flow_id]
        ]


class FlowProgress(BaseModel):
    data: dict[UUID, float] = Field(default_factory=dict)


class InfrahubEventFilter(EventFilter):
    def add_related_filter(self, related: EventRelatedFilter) -> None:
        if not isinstance(self.related, list):
            self.related = []

        self.related.append(related)

    def add_account_filter(self, account__ids: list[str] | None) -> None:
        if account__ids:
            self.add_related_filter(
                EventRelatedFilter(
                    labels=ResourceSpecification(
                        {"prefect.resource.role": "infrahub.account", "infrahub.resource.id": account__ids}
                    )
                )
            )

    def add_branch_filter(self, branches: list[str] | None = None) -> None:
        if branches:
            self.add_related_filter(
                EventRelatedFilter(
                    labels=ResourceSpecification(
                        {"prefect.resource.role": "infrahub.branch", "infrahub.resource.label": branches}
                    )
                )
            )

    def add_event_filter(self, level: int | None = None, has_children: bool | None = None) -> None:
        event_filter: dict[str, list[str] | str] = {}
        if level is not None:
            event_filter["infrahub.event.level"] = str(level)

        if has_children is not None:
            event_filter["infrahub.event.has_children"] = str(has_children).lower()

        if event_filter:
            event_filter["prefect.resource.role"] = "infrahub.event"
            self.add_related_filter(EventRelatedFilter(labels=ResourceSpecification(event_filter)))

    def add_event_id_filter(self, ids: list[str] | None = None) -> None:
        if ids:
            self.id = EventIDFilter(id=[uuid.UUID(id) for id in ids])

    def add_event_type_filter(
        self, event_type: list[str] | None = None, event_type_filter: dict[str, Any] | None = None
    ) -> None:
        event_type = event_type or []
        event_type_filter = event_type_filter or {}

        if branch_merged := event_type_filter.get("branch_merged"):
            branches: list[str] = branch_merged.get("branches") or []
            if "infrahub.branch.created" not in event_type:
                event_type.append("infrahub.branch.merged")
            if branches:
                self.resource = EventResourceFilter(labels=ResourceSpecification({"infrahub.branch.name": branches}))

        if branch_migrated := event_type_filter.get("branch_migrated"):
            branches = branch_migrated.get("branches") or []
            if "infrahub.branch.created" not in event_type:
                event_type.append("infrahub.branch.migrated")
            if branches:
                self.resource = EventResourceFilter(labels=ResourceSpecification({"infrahub.branch.name": branches}))

        if branch_rebased := event_type_filter.get("branch_rebased"):
            branches = branch_rebased.get("branches") or []
            if "infrahub.branch.created" not in event_type:
                event_type.append("infrahub.branch.rebased")
            if branches:
                self.resource = EventResourceFilter(labels=ResourceSpecification({"infrahub.branch.name": branches}))

        if event_type:
            self.event = EventNameFilter(name=event_type)

    def add_primary_node_filter(self, primary_node__ids: list[str] | None) -> None:
        if primary_node__ids:
            self.resource = EventResourceFilter(labels=ResourceSpecification({"infrahub.node.id": primary_node__ids}))

    def add_parent_filter(self, parent__ids: list[str] | None) -> None:
        if parent__ids:
            self.add_related_filter(
                EventRelatedFilter(
                    labels=ResourceSpecification(
                        {"prefect.resource.role": "infrahub.child_event", "infrahub.event_parent.id": parent__ids}
                    )
                )
            )

    def add_related_node_filter(self, related_node__ids: list[str] | None) -> None:
        if related_node__ids:
            self.add_related_filter(
                EventRelatedFilter(
                    labels=ResourceSpecification(
                        {"prefect.resource.role": "infrahub.related.node", "prefect.resource.id": related_node__ids}
                    )
                )
            )

    @classmethod
    def from_filters(
        cls,
        order: EventSortOrder,
        ids: list[str] | None = None,
        account__ids: list[str] | None = None,
        related_node__ids: list[str] | None = None,
        parent__ids: list[str] | None = None,
        primary_node__ids: list[str] | None = None,
        event_type: list[str] | None = None,
        event_type_filter: dict[str, Any] | None = None,
        branches: list[str] | None = None,
        level: int | None = None,
        has_children: bool | None = None,
        since: datetime | None = None,
        until: datetime | None = None,
    ) -> InfrahubEventFilter:
        occurred_filter = {}
        if since:
            occurred_filter["since"] = Timestamp(since.isoformat()).to_datetime()

        if until:
            occurred_filter["until"] = Timestamp(until.isoformat()).to_datetime()

        if occurred_filter:
            filters = cls(occurred=EventOccurredFilter(**occurred_filter))
        else:
            filters = cls()

        match order:
            case EventSortOrder.ASC:
                filters.order = PrefectEventOrder.ASC
            case EventSortOrder.DESC:
                filters.order = PrefectEventOrder.DESC

        filters.add_event_filter(level=level, has_children=has_children)
        filters.add_event_id_filter(ids=ids)
        filters.add_event_type_filter(event_type=event_type, event_type_filter=event_type_filter)
        filters.add_branch_filter(branches=branches)
        filters.add_account_filter(account__ids=account__ids)
        filters.add_parent_filter(parent__ids=parent__ids)
        filters.add_primary_node_filter(primary_node__ids=primary_node__ids)
        filters.add_related_node_filter(related_node__ids=related_node__ids)

        return filters



from typing import Any, assert_never

from infrahub_sdk.convert_object_type import ConversionFieldInput, ConversionFieldValue

from infrahub.core.attribute import BaseAttribute
from infrahub.core.branch import Branch
from infrahub.core.branch.enums import BranchStatus
from infrahub.core.constants import BranchSupportType, RelationshipCardinality
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.node.create import create_node
from infrahub.core.query.relationship import GetAllPeersIds
from infrahub.core.query.resource_manager import PoolChangeReserved
from infrahub.core.relationship import RelationshipManager
from infrahub.core.schema import NodeSchema
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase
from infrahub.message_bus.messages import RefreshRegistryBranches
from infrahub.tasks.registry import update_branch_registry
from infrahub.workers.dependencies import get_message_bus


def _get_conversion_field_raw_value(conv_field_value: ConversionFieldValue) -> Any:
    if conv_field_value.attribute_value is not None:
        return conv_field_value.attribute_value
    if conv_field_value.peer_id is not None:
        return conv_field_value.peer_id
    if conv_field_value.peers_ids is not None:
        return conv_field_value.peers_ids
    raise ValueError("ConversionFieldValue has not been validated correctly.")


async def get_out_rels_peers_ids(node: Node, db: InfrahubDatabase, at: Timestamp) -> list[str]:
    all_peers_ids: list[str] = []
    for name in node._relationships:
        relm: RelationshipManager = getattr(node, name)
        peers = await relm.get_db_peers(db=db, at=at)
        all_peers_ids.extend([str(peer.peer_id) for peer in peers])
    return all_peers_ids


async def build_data_new_node(db: InfrahubDatabase, mapping: dict[str, ConversionFieldInput], node: Node) -> dict:
    """Value of a given field on the target kind to convert is either an input source attribute/relationship of the source node,
    or a raw value."""

    data = {}
    for dest_field_name, conv_field_input in mapping.items():
        if conv_field_input.source_field is not None:
            # Fetch the value of the corresponding field from the node being converted.
            item = getattr(node, conv_field_input.source_field)
            if isinstance(item, BaseAttribute):
                data[dest_field_name] = item.value
            elif isinstance(item, RelationshipManager):
                if item.schema.cardinality == RelationshipCardinality.ONE:
                    peer = await item.get_peer(db=db)
                    if peer is not None:
                        data[dest_field_name] = {"id": peer.id}
                    # else, relationship is optional, and if the target relationship is mandatory an error will be raised during creation
                elif item.schema.cardinality == RelationshipCardinality.MANY:
                    data[dest_field_name] = [{"id": peer.id} for _, peer in (await item.get_peers(db=db)).items()]
                else:
                    assert_never(item.schema.cardinality)
        elif conv_field_input.data is not None:
            data[dest_field_name] = _get_conversion_field_raw_value(conv_field_input.data)
        elif conv_field_input.use_default_value is True:
            pass  # default value will be used automatically when creating the node
        else:
            raise ValueError("ConversionFieldInput has not been validated correctly.")
    return data


async def get_unidirectional_rels_peers_ids(
    node: Node, branch: Branch, db: InfrahubDatabase, at: Timestamp
) -> list[str]:
    """
    Returns peers ids of nodes connected to input `node` through an incoming unidirectional relationship.
    """

    out_rels_identifier = [rel.identifier for rel in node.get_schema().relationships]
    branch_agnostic = node.get_schema().branch == BranchSupportType.AGNOSTIC
    query = await GetAllPeersIds.init(
        db=db,
        node_id=node.id,
        branch=branch,
        exclude_identifiers=out_rels_identifier,
        branch_agnostic=branch_agnostic,
        at=at,
    )
    await query.execute(db=db)
    return query.get_peers_uuids()


async def _get_other_active_branches(db: InfrahubDatabase) -> list[Branch]:
    branches = await Branch.get_list(db=db)
    return [branch for branch in branches if not (branch.is_global or branch.is_default)]


def _has_pass_thru_aware_attributes(node_schema: NodeSchema, mapping: dict[str, ConversionFieldInput]) -> bool:
    aware_attributes = [attr for attr in node_schema.attributes if attr.branch != BranchSupportType.AGNOSTIC]
    aware_attributes_pass_thru = [
        attr.name for attr in aware_attributes if attr.name in mapping and mapping[attr.name].source_field is not None
    ]
    return len(aware_attributes_pass_thru) > 0


async def validate_conversion(
    deleted_node: Node, branch: Branch, db: InfrahubDatabase, timestamp_before_conversion: Timestamp
) -> None:
    deleted_node_out_rels_peer_ids = await get_out_rels_peers_ids(
        node=deleted_node, db=db, at=timestamp_before_conversion
    )
    deleted_node_unidir_rels_peer_ids = await get_unidirectional_rels_peers_ids(
        node=deleted_node, db=db, branch=branch, at=timestamp_before_conversion
    )

    # Make sure relationships with constraints are not broken by retrieving them
    peers_ids = deleted_node_out_rels_peer_ids + deleted_node_unidir_rels_peer_ids
    peers = await NodeManager.get_many(ids=peers_ids, db=db, prefetch_relationships=True, branch=branch)
    for peer in peers.values():
        peer.validate_relationships()


async def convert_and_validate_object_type(
    node: Node,
    target_schema: NodeSchema,
    mapping: dict[str, ConversionFieldInput],
    branch: Branch,
    db: InfrahubDatabase,
) -> Node:
    async with db.start_transaction() as dbt:
        timestamp_before_conversion = Timestamp()
        new_node = await convert_object_type(
            node=node, target_schema=target_schema, mapping=mapping, branch=branch, db=dbt
        )
        await validate_conversion(
            deleted_node=node, branch=branch, db=dbt, timestamp_before_conversion=timestamp_before_conversion
        )

    # Refresh outside the transaction otherwise other workers would pull outdated branch objects.
    message_bus = await get_message_bus()
    await message_bus.send(RefreshRegistryBranches())

    return new_node


async def convert_object_type(
    node: Node,
    target_schema: NodeSchema,
    mapping: dict[str, ConversionFieldInput],
    branch: Branch,
    db: InfrahubDatabase,
) -> Node:
    """Delete the node and return the new created one. If creation fails, the node is not deleted, and raise an error.
    An extra check is performed on input node peers relationships to make sure they are still valid."""

    node_schema = node.get_schema()
    if not isinstance(node_schema, NodeSchema):
        raise ValueError(f"Only a node with a NodeSchema can be converted, got {type(node_schema)}")

    # Delete the node, so we delete relationships with peers as well, which might temporarily break cardinality constraints
    # but they should be restored when creating the new node.
    deleted_nodes = await NodeManager.delete(db=db, branch=branch, nodes=[node], cascade_delete=False)
    if len(deleted_nodes) != 1:
        raise ValueError(f"Deleted {len(deleted_nodes)} nodes instead of 1")

    data_new_node = await build_data_new_node(db, mapping, node)

    if node_schema.branch == BranchSupportType.AGNOSTIC and _has_pass_thru_aware_attributes(
        node_schema=node_schema, mapping=mapping
    ):
        if not branch.is_default:
            raise ValueError(
                f"Conversion of {node_schema.kind} is not allowed on branch {branch.name} because it is agnostic and has aware attributes"
            )

        # When converting an agnostic node with aware attributes, we need to put other branches in NEED_REBASE state
        # as aware attributes do not exist in other branches after conversion
        other_branches = await _get_other_active_branches(db=db)
        for br in other_branches:
            br.status = BranchStatus.NEED_REBASE
            await br.save(db=db)
            # Registry of other API workers are updated outside the transaction
            await update_branch_registry(db=db, branch=br)

    node_created = await create_node(
        data=data_new_node,
        db=db,
        branch=branch,
        schema=target_schema,
    )

    # If the node had some value reserved in any Pools / Resource Manager, we need to change the identifier of the reservation(s)
    query = await PoolChangeReserved.init(
        db=db,
        existing_identifier=node.get_id(),
        new_identifier=node_created.get_id(),
        branch=branch,
    )
    await query.execute(db=db)

    return node_created



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from infrahub.core.schema.schema_branch import SchemaBranch


class SchemaBranchValidator:
    @abstractmethod
    def check(self, schema_branch: SchemaBranch) -> None:
        raise NotImplementedError()



from __future__ import annotations

import inspect
import os
from abc import abstractmethod
from typing import TYPE_CHECKING, Any

from .operation import InfrahubOperation

if TYPE_CHECKING:
    from . import InfrahubClient
    from .node import InfrahubNode

INFRAHUB_TRANSFORM_VARIABLE_TO_IMPORT = "INFRAHUB_TRANSFORMS"


class InfrahubTransform(InfrahubOperation):
    name: str | None = None
    query: str
    timeout: int = 60

    def __init__(
        self,
        client: InfrahubClient,
        infrahub_node: type[InfrahubNode],
        convert_query_response: bool = False,
        branch: str = "",
        root_directory: str = "",
        server_url: str = "",
    ) -> None:
        super().__init__(
            client=client,
            infrahub_node=infrahub_node,
            convert_query_response=convert_query_response,
            branch=branch,
            root_directory=root_directory,
        )

        self.server_url = server_url or os.environ.get("INFRAHUB_URL", "http://127.0.0.1:8000")
        self._client = client

        if not self.name:
            self.name = self.__class__.__name__

        if not self.query:
            raise ValueError("A query must be provided")

    @property
    def client(self) -> InfrahubClient:
        return self._init_client

    @abstractmethod
    def transform(self, data: dict) -> Any:
        pass

    async def collect_data(self) -> dict:
        """Query the result of the GraphQL Query defined in self.query and return the result"""

        return await self.client.query_gql_query(name=self.query, branch_name=self.branch_name)

    async def run(self, data: dict | None = None) -> Any:
        """Execute the transformation after collecting the data from the GraphQL query.

        The result of the check is determined based on the presence or not of ERROR log messages.

        Args:
            data: The data on which to run the transform. Data will be queried from the API if not provided

        Returns: Transformed data
        """

        if not data:
            data = await self.collect_data()

        unpacked = data.get("data") or data
        await self.process_nodes(data=unpacked)

        if inspect.iscoroutinefunction(self.transform):
            return await self.transform(data=unpacked)

        return self.transform(data=unpacked)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from graphene import Boolean, Field, Int, List, NonNull, ObjectType, String
from infrahub_sdk.utils import extract_fields_first_node

from infrahub.core.constants import InfrahubKind, MetadataOptions
from infrahub.core.manager import NodeManager
from infrahub.core.protocols import CoreGenericAccount, CoreProposedChange
from infrahub.proposed_change.action_checker import ACTION_RULES, ActionRulesEvaluator

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.graphql.initialization import GraphqlContext


class ActionAvailability(ObjectType):
    action = Field(String, required=True, description="The action that a user may want to take on a proposed change")
    available = Field(Boolean, required=True, description="Tells if the action is available")
    unavailability_reason = Field(String, required=False, description="The reason why an action may be unavailable")


class ActionAvailabilityEdge(ObjectType):
    node = Field(ActionAvailability, required=True)


class AvailableActions(ObjectType):
    count = Field(Int, required=True, description="The number of available actions for the proposed change.")
    edges = Field(List(of_type=NonNull(ActionAvailabilityEdge), required=True), required=True)

    @staticmethod
    async def resolve(
        root: dict,  # noqa: ARG004
        info: GraphQLResolveInfo,
        proposed_change_id: str,
    ) -> dict:
        graphql_context: GraphqlContext = info.context
        proposed_change = await NodeManager.get_one(
            kind=InfrahubKind.PROPOSEDCHANGE,
            id=proposed_change_id,
            db=graphql_context.db,
            branch=graphql_context.branch,
            raise_on_error=True,
            include_metadata=MetadataOptions.CREATED_BY,
        )

        author_id = proposed_change._get_created_by()
        if not author_id:
            raise ValueError("Proposed change author not found")

        proposed_change_author = await NodeManager.get_one(
            db=graphql_context.db,
            id=author_id,
            kind=CoreGenericAccount,
            raise_on_error=True,
            branch=graphql_context.branch,
        )
        actions = await ActionRulesEvaluator(rules=ACTION_RULES).evaluate(
            proposed_change=cast("CoreProposedChange", proposed_change),
            graphql_context=graphql_context,
            proposed_change_author=proposed_change_author,
        )

        fields = await extract_fields_first_node(info=info)
        response: dict[str, Any] = {}

        if "count" in fields:
            response["count"] = len(actions)

        if edges := fields.get("edges"):
            node_fields = edges.get("node", {})

            nodes = []
            for action in actions:
                node = {}

                if "action" in node_fields:
                    node["action"] = action["action"]
                if "available" in node_fields:
                    node["available"] = action["available"]
                if "unavailability_reason" in node_fields:
                    node["unavailability_reason"] = action["unavailability_reason"]

                nodes.append({"node": node})

            response["edges"] = nodes

        return response


ProposedChangeAvailableActions = Field(
    AvailableActions, proposed_change_id=String(required=True), resolver=AvailableActions.resolve, required=True
)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence
from uuid import uuid4

import pydantic
from prefect import flow, get_run_logger
from prefect.client.schemas.objects import State  # noqa: TC002
from prefect.states import Completed, Failed

from infrahub import config, lock
from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.branch.enums import BranchStatus
from infrahub.core.changelog.diff import DiffChangelogCollector, MigrationTracker
from infrahub.core.constants import DiffAction, MutationAction
from infrahub.core.diff.coordinator import DiffCoordinator
from infrahub.core.diff.diff_locker import DiffLocker
from infrahub.core.diff.ipam_diff_parser import IpamDiffParser
from infrahub.core.diff.merger.merger import DiffMerger
from infrahub.core.diff.model.path import BranchTrackingId, EnrichedDiffRoot, EnrichedDiffRootMetadata
from infrahub.core.diff.models import RequestDiffUpdate
from infrahub.core.diff.repository.repository import DiffRepository
from infrahub.core.graph import GRAPH_VERSION
from infrahub.core.merge import BranchMerger
from infrahub.core.merge.merge_locker import MergeLocker
from infrahub.core.migrations.exceptions import MigrationFailureError
from infrahub.core.migrations.runner import MigrationRunner
from infrahub.core.schema.update_coordinator import MigrationExecutor, SchemaUpdateCoordinator
from infrahub.core.timestamp import Timestamp
from infrahub.core.validators.determiner import ConstraintValidatorDeterminer
from infrahub.core.validators.models.validate_migration import SchemaValidateMigrationData
from infrahub.core.validators.tasks import schema_validate_migrations
from infrahub.dependencies.registry import get_component_registry
from infrahub.events.branch_action import (
    BranchCreatedEvent,
    BranchDeletedEvent,
    BranchMergedEvent,
    BranchMigratedEvent,
    BranchRebasedEvent,
)
from infrahub.events.models import EventMeta, InfrahubEvent
from infrahub.events.node_action import get_node_event
from infrahub.exceptions import BranchNotFoundError, ValidationError
from infrahub.generators.constants import GeneratorDefinitionRunSource
from infrahub.graphql.mutations.models import BranchCreateModel  # noqa: TC001
from infrahub.workers.dependencies import get_component, get_database, get_event_service, get_workflow
from infrahub.workflows.catalogue import (
    BRANCH_CANCEL_PROPOSED_CHANGES,
    BRANCH_MERGE_POST_PROCESS,
    DIFF_REFRESH_ALL,
    DIFF_UPDATE,
    GIT_REPOSITORIES_CREATE_BRANCH,
    IPAM_RECONCILIATION,
    TRIGGER_ARTIFACT_DEFINITION_GENERATE,
    TRIGGER_GENERATOR_DEFINITION_RUN,
)
from infrahub.workflows.utils import add_tags

if TYPE_CHECKING:
    from logging import Logger, LoggerAdapter

    from infrahub.core.changelog.models import NodeChangelog
    from infrahub.database import InfrahubDatabase


@flow(name="branch-migrate", flow_run_name="Apply migrations to branch {branch}")
async def migrate_branch(branch: str, context: InfrahubContext, send_events: bool = True) -> None:
    await add_tags(branches=[branch])

    database = await get_database()
    async with database.start_session() as db:
        log = get_run_logger()

        obj = await Branch.get_by_name(db=db, name=branch)

        if obj.graph_version == GRAPH_VERSION:
            log.info(f"Branch '{obj.name}' has graph version {obj.graph_version}, no migrations to apply")
            return

        migration_runner = MigrationRunner(branch=obj)
        if not migration_runner.has_migrations():
            log.info(f"No migrations detected for branch '{obj.name}'")
            obj.graph_version = GRAPH_VERSION
            await obj.save(db=db)
            return

        # Branch status will remain as so if the migration process fails
        # This will help user to know that a branch is in an invalid state to be used properly and that actions need to be taken
        if obj.status != BranchStatus.NEED_UPGRADE_REBASE:
            obj.status = BranchStatus.NEED_UPGRADE_REBASE
            await obj.save(db=db)

        try:
            log.info(f"Running migrations for branch '{obj.name}'")
            await migration_runner.run(db=db, at=Timestamp())
        except MigrationFailureError as exc:
            log.error(f"Failed to run migrations for branch '{obj.name}': {exc.errors}")
            raise

        if obj.status == BranchStatus.NEED_UPGRADE_REBASE:
            obj.status = BranchStatus.OPEN
        obj.graph_version = GRAPH_VERSION
        await obj.save(db=db)

    if send_events:
        event_service = await get_event_service()
        await event_service.send(
            BranchMigratedEvent(
                branch_name=obj.name, branch_id=str(obj.uuid), meta=EventMeta(branch=obj, context=context)
            )
        )


@flow(name="branch-rebase", flow_run_name="Rebase branch {branch}")
async def rebase_branch(branch: str, context: InfrahubContext, send_events: bool = True) -> None:  # noqa: PLR0915
    workflow = get_workflow()
    database = await get_database()
    async with database.start_session() as db:
        log = get_run_logger()
        await add_tags(branches=[branch])
        obj = await Branch.get_by_name(db=db, name=branch)
        base_branch = await Branch.get_by_name(db=db, name=registry.default_branch)
        component_registry = get_component_registry()
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        diff_coordinator = await component_registry.get_component(DiffCoordinator, db=db, branch=obj)
        diff_coordinator.set_logger(log)
        diff_merger = await component_registry.get_component(DiffMerger, db=db, branch=obj)
        initial_from_time = Timestamp(obj.get_branched_from())
        merger = BranchMerger(
            db=db,
            diff_coordinator=diff_coordinator,
            diff_merger=diff_merger,
            diff_repository=diff_repository,
            source_branch=obj,
            diff_locker=DiffLocker(),
            workflow=workflow,
        )

        enriched_diff_metadata = await diff_coordinator.update_branch_diff(base_branch=base_branch, diff_branch=obj)
        async for _ in diff_repository.get_all_conflicts_for_diff(
            diff_branch_name=enriched_diff_metadata.diff_branch_name, diff_id=enriched_diff_metadata.uuid
        ):
            # if there are any conflicts, raise the error
            raise ValidationError(
                f"Branch {obj.name} contains conflicts with the default branch that must be addressed."
                " Please review the diff for details and manually update the conflicts before rebasing."
            )

        # rebase to the end time of the diff in case conflicting changes happen on
        # either branch while rebasing and migrating
        rebase_at = enriched_diff_metadata.to_time
        node_diff_field_summaries = await diff_repository.get_node_field_summaries(
            diff_branch_name=enriched_diff_metadata.diff_branch_name, diff_id=enriched_diff_metadata.uuid
        )

        candidate_schema = merger.get_candidate_schema()
        determiner = ConstraintValidatorDeterminer(schema_branch=candidate_schema)
        constraints = await determiner.get_constraints(node_diffs=node_diff_field_summaries)

        # If there are some changes related to the schema between this branch and main, we need to
        #  - Run all the validations to ensure everything is correct before rebasing the branch
        #  - Run all the migrations after the rebase
        if obj.has_schema_changes:
            constraints += await merger.calculate_validations(target_schema=candidate_schema)
        if constraints:
            responses = await schema_validate_migrations(
                message=SchemaValidateMigrationData(branch=obj, schema_branch=candidate_schema, constraints=constraints)
            )
            error_messages = [
                f"{violation.message} for constraint {response.constraint_name} {response.schema_path.field_name} {response.schema_path.property_name} and node {violation.node_id} {violation.node_kind}"  # noqa: E501
                for response in responses
                for violation in response.violations
            ]

            if error_messages:
                raise ValidationError(",\n".join(error_messages))

        pre_rebase_schema = merger.destination_schema.duplicate()
        migrations = []
        async with lock.registry.global_graph_lock():
            async with db.start_transaction() as dbt:
                await obj.rebase(db=dbt, user_id=context.account.account_id, at=rebase_at)
                log.info("Branch graph rebased")

            if obj.has_schema_changes:
                # Load the updated schema from DB after rebase
                log.info("Loading rebased schema")
                updated_schema = await registry.schema.load_schema_from_db(db=db, branch=obj)

                # Calculate migrations before updating registry
                log.info("Calculating migrations")
                migrations = await merger.calculate_migrations(target_schema=updated_schema)

                # Use coordinator to update registry and run migrations with rollback on failure
                log.info("Running migrations")
                coordinator = SchemaUpdateCoordinator(
                    db=db,
                    branch=obj,
                    schema_manager=registry.schema,
                    origin_schema=pre_rebase_schema,
                    workflow=workflow,
                    context=context,
                    migration_executor=MigrationExecutor.WORKFLOW if send_events else MigrationExecutor.DIRECT,
                    logger=log,
                )
                await coordinator.execute(
                    candidate_schema=updated_schema,
                    at=rebase_at,
                    migrations=migrations,
                    update_db=False,
                    update_registry=True,
                    user_id=context.account.account_id,
                )
                log.info("Migrations completed")

        default_branch_diff = await _get_diff_root(
            diff_coordinator=diff_coordinator,
            enriched_diff_metadata=enriched_diff_metadata,
            diff_repository=diff_repository,
            base_branch=base_branch,
            target_from=initial_from_time,
        )

        # -------------------------------------------------------------
        # Trigger the reconciliation of IPAM data after the rebase
        # -------------------------------------------------------------
        diff_parser = await component_registry.get_component(IpamDiffParser, db=db, branch=obj)
        ipam_node_details = await diff_parser.get_changed_ipam_node_details(
            source_branch_name=obj.name,
            target_branch_name=registry.default_branch,
        )
        if ipam_node_details:
            await workflow.submit_workflow(
                workflow=IPAM_RECONCILIATION,
                context=context,
                parameters={"branch": obj.name, "ipam_node_details": ipam_node_details},
            )

    await migrate_branch(branch=branch, context=context, send_events=send_events)
    await workflow.submit_workflow(workflow=DIFF_REFRESH_ALL, context=context, parameters={"branch_name": obj.name})

    if not send_events:
        return

    # -------------------------------------------------------------
    # Generate an event to indicate that a branch has been rebased
    # -------------------------------------------------------------
    rebase_event = BranchRebasedEvent(
        branch_name=obj.name, branch_id=str(obj.uuid), meta=EventMeta(branch=obj, context=context)
    )
    events: list[InfrahubEvent] = [rebase_event]
    changelog_collector = DiffChangelogCollector(
        diff=default_branch_diff, branch=obj, db=db, migration_tracker=MigrationTracker(migrations=migrations)
    )
    for action, node_changelog in changelog_collector.collect_changelogs():
        node_event_class = get_node_event(MutationAction.from_diff_action(diff_action=action))
        mutate_event = node_event_class(
            kind=node_changelog.node_kind,
            node_id=node_changelog.node_id,
            changelog=node_changelog,
            fields=node_changelog.updated_fields,
            meta=EventMeta.from_parent(parent=rebase_event, branch=obj),
        )
        events.append(mutate_event)

    event_service = await get_event_service()
    for event in events:
        await event_service.send(event)


@flow(name="branch-merge", flow_run_name="Merge branch {branch} into main")
async def merge_branch(branch: str, context: InfrahubContext, proposed_change_id: str | None = None) -> None:
    database = await get_database()
    async with database.start_session() as db:
        log = get_run_logger()

        await add_tags(branches=[branch, registry.default_branch])

        obj = await Branch.get_by_name(db=db, name=branch)
        default_branch = await registry.get_branch(db=db, branch=registry.default_branch)
        merge_event = BranchMergedEvent(
            branch_name=obj.name,
            branch_id=str(obj.get_uuid()),
            proposed_change_id=proposed_change_id,
            meta=EventMeta.from_context(context=context, branch=registry.get_global_branch()),
        )

        merge_locker = MergeLocker()
        async with merge_locker.acquire_global_lock():
            obj = await Branch.get_by_name(db=db, name=branch)
            if obj.status != BranchStatus.OPEN:
                log.info(f"Branch '{branch}' is not open (status={obj.status}), skipping merge")
                return

            node_events = await _do_merge_branch(
                db=db,
                log=log,
                obj=obj,
                context=context,
            )

        events: list[InfrahubEvent] = [merge_event]

        for action, node_changelog in node_events:
            meta = EventMeta.from_parent(parent=merge_event, branch=default_branch)
            node_event_class = get_node_event(MutationAction.from_diff_action(diff_action=action))
            mutate_event = node_event_class(
                kind=node_changelog.node_kind,
                node_id=node_changelog.node_id,
                changelog=node_changelog,
                fields=node_changelog.updated_fields,
                meta=meta,
            )
            events.append(mutate_event)

        event_service = await get_event_service()
        for event in events:
            await event_service.send(event=event)


async def _do_merge_branch(
    db: InfrahubDatabase, log: Logger | LoggerAdapter, obj: Branch, context: InfrahubContext
) -> Sequence[tuple[DiffAction, NodeChangelog]]:
    component_registry = get_component_registry()

    merger: BranchMerger | None = None
    workflow = get_workflow()
    merge_at = Timestamp()
    pre_merge_schema = registry.schema.get_schema_branch(name=registry.default_branch).duplicate()
    async with lock.registry.global_graph_lock():
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        diff_coordinator = await component_registry.get_component(DiffCoordinator, db=db, branch=obj)
        diff_merger = await component_registry.get_component(DiffMerger, db=db, branch=obj)
        merger = BranchMerger(
            db=db,
            diff_coordinator=diff_coordinator,
            diff_merger=diff_merger,
            diff_repository=diff_repository,
            source_branch=obj,
            diff_locker=DiffLocker(),
            workflow=workflow,
        )
        branch_diff = await merger.merge(at=merge_at)

    changelog_collector = DiffChangelogCollector(diff=branch_diff, branch=obj, db=db)
    node_events = changelog_collector.collect_changelogs()

    # Handle schema updates and migrations after merge
    if merger and await merger.has_schema_changes():
        # Load the updated schema from DB after merge
        log.info("Loading updated schema")
        updated_schema = await registry.schema.load_schema_from_db(
            db=db,
            branch=merger.destination_branch,
        )
        log.info("Calculating migrations")
        migrations = await merger.calculate_migrations(target_schema=updated_schema)

        # Use coordinator to update registry and run migrations with rollback on failure
        log.info("Running migrations")
        coordinator = SchemaUpdateCoordinator(
            db=db,
            branch=merger.destination_branch,
            schema_manager=registry.schema,
            origin_schema=pre_merge_schema,
            workflow=workflow,
            context=context,
            migration_executor=MigrationExecutor.WORKFLOW,
            logger=log,
        )
        await coordinator.execute(
            candidate_schema=updated_schema,
            at=merge_at,
            migrations=migrations,
            update_db=False,  # Schema nodes already written by merge
            update_registry=True,
            user_id=context.account.account_id,
        )
        log.info("Migrations completed")
    # -------------------------------------------------------------
    # Trigger the reconciliation of IPAM data after the merge
    # -------------------------------------------------------------
    diff_parser = await component_registry.get_component(IpamDiffParser, db=db, branch=obj)
    ipam_node_details = await diff_parser.get_changed_ipam_node_details(
        source_branch_name=obj.name,
        target_branch_name=registry.default_branch,
    )
    if ipam_node_details:
        await workflow.submit_workflow(
            workflow=IPAM_RECONCILIATION,
            context=context,
            parameters={"branch": registry.default_branch, "ipam_node_details": ipam_node_details},
        )
    # -------------------------------------------------------------
    # remove tracking ID from the diff because there is no diff after the merge
    # -------------------------------------------------------------
    await diff_repository.mark_tracking_ids_merged(tracking_ids=[BranchTrackingId(name=obj.name)])
    await diff_repository.freeze_diffs_for_branch(branch_name=obj.name)

    # -------------------------------------------------------------
    # Set branch status to MERGED to make it read-only
    # -------------------------------------------------------------
    obj.status = BranchStatus.MERGED
    await obj.save(db=db)
    registry.branch[obj.name] = obj

    # -------------------------------------------------------------
    # Cancel any remaining open proposed changes for this merged branch
    # -------------------------------------------------------------
    await workflow.submit_workflow(
        workflow=BRANCH_CANCEL_PROPOSED_CHANGES,
        context=context,
        parameters={"branch_name": obj.name},
    )

    # -------------------------------------------------------------
    # Generate an event to indicate that a branch has been merged
    # NOTE: we still need to convert this event and potentially pull
    #   some tasks currently executed based on the event into this workflow
    # -------------------------------------------------------------
    await workflow.submit_workflow(
        workflow=BRANCH_MERGE_POST_PROCESS,
        context=context,
        parameters={"source_branch": obj.name, "target_branch": registry.default_branch},
    )

    return node_events


@flow(name="branch-delete", flow_run_name="Delete branch {branch}")
async def delete_branch(branch: str, context: InfrahubContext) -> None:
    await add_tags(branches=[branch])

    database = await get_database()
    async with database.start_session() as db:
        obj = await Branch.get_by_name(db=db, name=str(branch))

        component_registry = get_component_registry()
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        await diff_repository.freeze_diffs_for_branch(branch_name=branch)

        await obj.delete(db=db)

        event = BranchDeletedEvent(
            branch_name=branch,
            branch_id=str(obj.uuid),
            sync_with_git=obj.sync_with_git,
            meta=EventMeta.from_context(context=context, branch=registry.get_global_branch()),
        )

        await get_workflow().submit_workflow(
            workflow=BRANCH_CANCEL_PROPOSED_CHANGES, context=context, parameters={"branch_name": branch}
        )

        event_service = await get_event_service()
        await event_service.send(event=event)


@flow(
    name="branch-validate",
    flow_run_name="Validate branch {branch} for conflicts",
    description="Validate if the branch has some conflicts",
    persist_result=True,
)
async def validate_branch(branch: str) -> State:
    await add_tags(branches=[branch])

    database = await get_database()
    async with database.start_session() as db:
        obj = await Branch.get_by_name(db=db, name=branch)

        component_registry = get_component_registry()
        diff_repo = await component_registry.get_component(DiffRepository, db=db, branch=obj)
        has_conflicts = await diff_repo.diff_has_conflicts(
            diff_branch_name=obj.name, tracking_id=BranchTrackingId(name=obj.name)
        )
        if has_conflicts:
            return Failed(message="branch has some conflicts")
        return Completed(message="branch is valid")


@flow(name="create-branch", flow_run_name="Create branch {model.name}")
async def create_branch(model: BranchCreateModel, context: InfrahubContext) -> None:
    await add_tags(branches=[model.name])

    database = await get_database()
    async with database.start_session() as db:
        try:
            await Branch.get_by_name(db=db, name=model.name)
            raise ValidationError(f"The branch {model.name} already exists")
        except BranchNotFoundError:
            pass

        data_dict: dict[str, Any] = dict(model)
        data_dict.pop("is_isolated", None)

        try:
            obj = Branch(**data_dict)
        except pydantic.ValidationError as exc:
            error_msgs = [f"invalid field {error['loc'][0]}: {error['msg']}" for error in exc.errors()]
            raise ValidationError("\n".join(error_msgs)) from exc

        async with lock.registry.local_schema_lock():
            # Copy the schema from the origin branch and set the hash and the schema_changed_at value
            origin_schema = registry.schema.get_schema_branch(name=obj.origin_branch)
            new_schema = origin_schema.duplicate(name=obj.name)
            registry.schema.set_schema_branch(name=obj.name, schema=new_schema)
            obj.update_schema_hash()
            await obj.save(db=db, user_id=context.account.account_id)

            # Add Branch to registry
            registry.branch[obj.name] = obj
            component = await get_component()
            await component.refresh_schema_hash(branches=[obj.name])

        event = BranchCreatedEvent(
            branch_name=obj.name,
            branch_id=str(obj.uuid),
            sync_with_git=obj.sync_with_git,
            meta=EventMeta.from_context(context=context, branch=registry.get_global_branch()),
        )
        event_service = await get_event_service()
        await event_service.send(event=event)

        if obj.sync_with_git:
            await get_workflow().submit_workflow(
                workflow=GIT_REPOSITORIES_CREATE_BRANCH,
                context=context,
                parameters={"branch": obj.name, "branch_id": str(obj.uuid)},
            )


async def _get_diff_root(
    diff_coordinator: DiffCoordinator,
    enriched_diff_metadata: EnrichedDiffRootMetadata,
    diff_repository: DiffRepository,
    base_branch: Branch,
    target_from: Timestamp,
) -> EnrichedDiffRoot:
    default_branch_diff = await diff_coordinator.create_or_update_arbitrary_timeframe_diff(
        base_branch=base_branch,
        diff_branch=base_branch,
        from_time=target_from,
        to_time=enriched_diff_metadata.to_time,
        name=str(uuid4()),
    )
    # make sure we have the actual diff with data and not just the metadata
    if not isinstance(default_branch_diff, EnrichedDiffRoot):
        default_branch_diff = await diff_repository.get_one(
            diff_branch_name=base_branch.name, diff_id=default_branch_diff.uuid
        )

    return default_branch_diff


@flow(
    name="branch-merge-post-process",
    flow_run_name="Run additional tasks after merging {source_branch} in {target_branch}",
)
async def post_process_branch_merge(source_branch: str, target_branch: str, context: InfrahubContext) -> None:
    database = await get_database()
    async with database.start_session() as db:
        await add_tags(branches=[source_branch])
        log = get_run_logger()
        log.info(f"Running additional tasks after merging {source_branch} within {target_branch}")

        component_registry = get_component_registry()
        default_branch = registry.get_branch_from_registry()
        diff_repository = await component_registry.get_component(DiffRepository, db=db, branch=default_branch)

        await get_workflow().submit_workflow(
            workflow=TRIGGER_ARTIFACT_DEFINITION_GENERATE,
            context=context,
            parameters={"branch": target_branch},
        )

        await get_workflow().submit_workflow(
            workflow=TRIGGER_GENERATOR_DEFINITION_RUN,
            context=context,
            parameters={"branch": target_branch, "source": GeneratorDefinitionRunSource.MERGE},
        )

        if not config.SETTINGS.main.diff_update_after_merge:
            return

        # send diff update requests for every active branch-tracking diff
        active_branches = await Branch.get_list(db=db)
        active_branch_names = {branch.name for branch in active_branches}
        diff_roots = await diff_repository.get_roots_metadata(base_branch_names=[target_branch])
        for diff_root in diff_roots:
            if (
                diff_root.base_branch_name != diff_root.diff_branch_name
                and diff_root.diff_branch_name in active_branch_names
                and isinstance(diff_root.tracking_id, BranchTrackingId)
                and not diff_root.is_frozen
            ):
                request_diff_update_model = RequestDiffUpdate(branch_name=diff_root.diff_branch_name)
                await get_workflow().submit_workflow(
                    workflow=DIFF_UPDATE, context=context, parameters={"model": request_diff_update_model}
                )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core import registry
from infrahub.core.constants import GLOBAL_BRANCH_NAME, BranchSupportType
from infrahub.core.query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.core.schema import AttributeSchema
    from infrahub.database import InfrahubDatabase


class UpdateAttributeValuesQuery(Query):
    """
    Update the values of the given attribute schema for the input node-id-to-value map.

    This version only expires existing values when they're different from the new value,
    making it safe to run idempotently without clearing correct existing values.

    This code is adapted from m044_backfill_hfid_display_label_in_db.
    """

    name = "update_attribute_values"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, attribute_schema: AttributeSchema, values_by_id_map: dict[str, Any], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.attribute_name = attribute_schema.name
        self.is_branch_agnostic = attribute_schema.get_branch() is BranchSupportType.AGNOSTIC
        self.values_by_id_map = values_by_id_map

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params = {
            "node_uuids": list(self.values_by_id_map.keys()),
            "attribute_name": self.attribute_name,
            "values_by_id": self.values_by_id_map,
            "default_branch": registry.default_branch,
            "global_branch": GLOBAL_BRANCH_NAME,
            "branch": GLOBAL_BRANCH_NAME if self.is_branch_agnostic else self.branch.name,
            "branch_level": 1 if self.is_branch_agnostic else self.branch.hierarchy_level,
            "at": self.at.to_string(),
        }
        branch_filter, branch_filter_params = self.branch.get_query_filter_path(at=self.at)
        self.params.update(branch_filter_params)

        if self.branch.name in [registry.default_branch, GLOBAL_BRANCH_NAME]:
            update_value_query = """
// ------------
// Find the Nodes and Attributes we need to update
// ------------
MATCH (n:Node)-[e:IS_PART_OF]->(:Root)
WHERE n.uuid IN $node_uuids
AND e.branch IN [$default_branch, $global_branch]
AND e.to IS NULL
AND e.status = "active"
WITH DISTINCT n
MATCH (n)-[e:HAS_ATTRIBUTE]->(attr:Attribute {name: $attribute_name})
WHERE e.branch IN [$default_branch, $global_branch]
AND e.to IS NULL
AND e.status = "active"
// ------------
// If the attribute has an existing value on the branch, then set the to time on it
// but only if the value is different from the new value
// ------------
WITH DISTINCT n, attr
CALL (attr) {
    OPTIONAL MATCH (attr)-[e:HAS_VALUE]->(existing_av)
    WHERE e.branch IN [$default_branch, $global_branch]
    AND e.to IS NULL
    AND e.status = "active"
    RETURN existing_av, e AS existing_has_value
}
CALL (existing_has_value, existing_av, n) {
    WITH existing_has_value, existing_av, n
    WHERE existing_has_value IS NOT NULL
    AND existing_av.value <> $values_by_id[n.uuid]
    SET existing_has_value.to = $at
}
WITH n, attr, existing_av
            """
        else:
            update_value_query = """
// ------------
// Find the Nodes and Attributes we need to update
// ------------
MATCH (n:Node)
WHERE n.uuid IN $node_uuids
CALL (n) {
    MATCH (n)-[r:IS_PART_OF]->(:Root)
    WHERE %(branch_filter)s
    RETURN r.status = "active" AS is_active
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
}
WITH n, is_active
WHERE is_active = TRUE
WITH DISTINCT n
CALL (n) {
    MATCH (n)-[r:HAS_ATTRIBUTE]->(attr:Attribute {name: $attribute_name})
    WHERE %(branch_filter)s
    RETURN attr, r.status = "active"  AS is_active
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
}
WITH DISTINCT n, attr, is_active
WHERE is_active = TRUE
// ------------
// If the attribute has an existing value on the branch, then set the to time on it
// but only if the value is different from the new value
// ------------
CALL (n, attr) {
    OPTIONAL MATCH (attr)-[r:HAS_VALUE]->(existing_av)
    WHERE %(branch_filter)s
    WITH r, existing_av, n
    ORDER BY r.branch_level DESC, r.from DESC, r.status ASC
    LIMIT 1
    WITH CASE
        WHEN existing_av.value <> $values_by_id[n.uuid]
        AND r.status = "active"
        AND r.branch = $branch
        THEN [r, existing_av]
        ELSE [NULL, NULL]
    END AS existing_details
    RETURN existing_details[0] AS existing_has_value, existing_details[1] AS existing_av
}
CALL (existing_has_value) {
    WITH existing_has_value
    WHERE existing_has_value IS NOT NULL
    SET existing_has_value.to = $at
}
WITH n, attr, existing_av
            """ % {"branch_filter": branch_filter}
        self.add_to_query(update_value_query)

        set_value_query = """
// ------------
// only make updates if the existing value is not the same as the new value
// ------------
WITH n, attr, existing_av, $values_by_id[n.uuid] AS required_value
WHERE existing_av.value <> required_value
OR existing_av IS NULL
CALL (n, attr) {
    MERGE (av:AttributeValue&AttributeValueIndexed {is_default: false, value: $values_by_id[n.uuid]} )
    WITH av, attr
    LIMIT 1
    CREATE (attr)-[r:HAS_VALUE { branch: $branch, branch_level: $branch_level, status: "active", from: $at }]->(av)
}
            """
        self.add_to_query(set_value_query)



from __future__ import annotations

from typing import TYPE_CHECKING, Self

from infrahub.core.registry import registry
from infrahub.events import NodeUpdatedEvent
from infrahub.trigger.constants import NAME_SEPARATOR
from infrahub.trigger.models import EventTrigger, ExecuteWorkflow, TriggerBranchDefinition, TriggerType

if TYPE_CHECKING:
    from infrahub.workflows.models import WorkflowDefinition


class ProfileRefreshTriggerDefinition(TriggerBranchDefinition):
    """Trigger definition for profile refresh when profile attributes/relationships change."""

    type: TriggerType = TriggerType.PROFILE
    profile_kind: str

    @classmethod
    def from_profile_schema(
        cls,
        branch: str,
        profile_kind: str,
        trigger_fields: list[str],
        workflow: WorkflowDefinition,
        branches_out_of_scope: list[str] | None = None,
    ) -> Self:
        """Create a trigger definition for profile refresh when profile attributes/relationships change."""
        event_trigger = EventTrigger()
        event_trigger.events.add(NodeUpdatedEvent.event_name)
        event_trigger.match = {"infrahub.node.kind": profile_kind}

        if branches_out_of_scope:
            event_trigger.match["infrahub.branch.name"] = [f"!{b}" for b in branches_out_of_scope]
        elif branch != registry.default_branch:
            event_trigger.match["infrahub.branch.name"] = branch

        event_trigger.match_related = {
            "prefect.resource.role": ["infrahub.node.attribute_update", "infrahub.node.relationship_update"],
            "infrahub.field.name": trigger_fields,
        }

        workflow_action = ExecuteWorkflow(
            workflow=workflow,
            parameters={
                "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                "profile_kind": profile_kind,
                "profile_id": "{{ event.resource['infrahub.node.id'] }}",
                "context": {
                    "__prefect_kind": "json",
                    "value": {
                        "__prefect_kind": "jinja",
                        "template": "{{ event.payload['context'] | tojson }}",
                    },
                },
            },
        )

        return cls(
            name=f"{profile_kind}{NAME_SEPARATOR}refresh",
            branch=branch,
            profile_kind=profile_kind,
            trigger=event_trigger,
            actions=[workflow_action],
        )



from __future__ import annotations

import contextlib
import hashlib
import io
import mimetypes
from dataclasses import dataclass
from typing import TYPE_CHECKING

import puremagic
from infrahub_sdk.uuidt import UUIDT

from infrahub import config
from infrahub.core import registry
from infrahub.core.constants import InfrahubKind
from infrahub.exceptions import ValidationError

if TYPE_CHECKING:
    from starlette.datastructures import UploadFile

    from infrahub.core.node import Node


@dataclass
class FileMetadata:
    """Metadata extracted from an uploaded file."""

    file_name: str
    checksum: str
    file_size: int
    file_type: str


@dataclass
class FileUploadResult:
    """Result of a file upload including storage identifier."""

    storage_id: str
    metadata: FileMetadata


class FileUploadProcessor:
    """Processor for handling file uploads."""

    def __init__(self, file: UploadFile) -> None:
        self.file = file
        self._metadata: FileMetadata | None = None
        self.storage_id: str | None = None

    @staticmethod
    def _format_file_size(size_bytes: int) -> str:
        """Format a file size in bytes to a human-readable string.

        Args:
            size_bytes: The file size in bytes.

        Returns:
            A human-readable string like "1.5 MB" or "256 KB".
        """
        value = float(size_bytes)
        for unit in ("B", "KB", "MB", "GB", "TB"):
            if abs(value) < 1024:
                return f"{value:.1f} {unit}" if unit != "B" else f"{int(value)} {unit}"
            value /= 1024
        return f"{value:.1f} PB"

    @staticmethod
    def _looks_like_text(content: bytes) -> bool:
        """Heuristic to determine if file content looks like text.

        https://github.com/file/file/blob/f2a6e7cb7db9b5fd86100403df6b2f830c7f22ba/src/encoding.c#L151-L228

        Args:
            content: The initial bytes of the file content.

        Returns:
            True if the content looks like text, False otherwise.
        """
        if b"\x00" in content:
            return False

        textchars = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7F})
        return not content.translate(None, textchars)

    @staticmethod
    def _detect_mime_type(content: bytes, filename: str | None = None) -> str:
        """Detect the MIME type of a file using magic bytes and optional filename.

        Passes the filename to puremagic so it can use the extension to improve detection
        of text-based formats (YAML, TOML, etc.) that lack distinctive magic bytes.

        Falls back to `application/octet-stream` if the type cannot be determined.

        Args:
            content: The file content (or first few KB) as bytes.
            filename: Optional original filename used to refine detection.

        Returns:
            The detected MIME type string.
        """
        with contextlib.suppress(puremagic.PureError):
            results = puremagic.magic_string(content, filename=filename)
            if results:
                return results[0].mime_type

        if filename and FileUploadProcessor._looks_like_text(content=content):
            # puremagic does not recognize some text-based format such as CSV
            guessed_type, _ = mimetypes.guess_type(filename)
            if guessed_type:
                return guessed_type

        return "application/octet-stream"

    @property
    def metadata(self) -> FileMetadata:
        """Get the extracted file metadata. Mostly to make type checkers happy.

        Returns:
            The extracted file metadata.

        Raises:
            RuntimeError: If metadata has not been extracted yet.
        """
        if not self._metadata:
            raise RuntimeError("File metadata has not been extracted yet.")

        return self._metadata

    def _get_file_size(self) -> int:
        """Get the size of the uploaded file without loading it into memory.

        Returns:
            The file size in bytes.
        """
        # Access underlying file object which supports seek with whence parameter
        # Starlette's UploadFile.seek() only accepts offset, not whence
        self.file.file.seek(0, io.SEEK_END)
        size = self.file.file.tell()
        self.file.file.seek(0)
        return size

    async def _compute_checksum(self) -> str:
        """Compute SHA-1 checksum of the file using chunked reading.

        Reads the file in 64KB chunks to avoid loading large files entirely into memory.

        Returns:
            The checksum as a hex string.
        """
        await self.file.seek(0)
        hasher = hashlib.sha1(usedforsecurity=False)

        while chunk := await self.file.read(65536):
            hasher.update(chunk)

        return hasher.hexdigest()

    async def _extract_metadata(self) -> None:
        """Extract and validate file metadata without storing the file.

        Raises:
            ValidationError: If the file exceeds the maximum allowed size.
        """
        if self._metadata:
            return

        file_size = self._get_file_size()
        if file_size > config.SETTINGS.storage.max_file_size * 1024 * 1024:
            raise ValidationError(
                f"File size ({self._format_file_size(size_bytes=file_size)}) exceeds maximum allowed size "
                f"({config.SETTINGS.storage.max_file_size} MB)"
            )

        magic_bytes = await self.file.read(2048)
        file_name = self.file.filename or ""
        file_type = self._detect_mime_type(content=magic_bytes, filename=file_name or None)
        checksum = await self._compute_checksum()

        self._metadata = FileMetadata(file_name=file_name, checksum=checksum, file_size=file_size, file_type=file_type)

    def _should_store_file(self, node: Node | None) -> bool:
        """Determine if the file should be stored based on checksum comparison.

        For idempotent upsert operations, compares the uploaded file's checksum
        with the existing node's checksum. If they match, storage can be skipped.

        Args:
            node: The existing node to compare checksums with, or `None` if creating new.

        Returns:
            True if the file should be stored, False if storage can be skipped.
        """
        # Skip storage only if node exists, is a FileObject, and checksums match
        return not (
            node
            and InfrahubKind.FILEOBJECT in node.get_schema().inherit_from
            and node.checksum.value == self.metadata.checksum  # type: ignore
        )

    async def _store_file(self) -> FileUploadResult:
        """Store the file in the storage backend.

        Returns:
            FileUploadResult containing all file metadata including storage_id.
        """
        self.storage_id = str(UUIDT())
        # Update file_name to use storage_id if original was unnamed
        self.metadata.file_name = self.file.filename or self.storage_id

        await self.file.seek(0)
        registry.storage.store(identifier=self.storage_id, content=self.file.file)

        return FileUploadResult(storage_id=self.storage_id, metadata=self.metadata)

    async def process(self, node: Node | None = None) -> FileUploadResult | None:
        """Process the file upload and store it in the storage backend.

        For idempotent operations (upsert), pass the existing node to compare checksums.
        If the checksums match, storage is skipped and `None` is returned.

        Args:
            node: Optional existing node to compare checksums with.
                  If provided and checksums match, storage is skipped.

        Returns:
            FileUploadResult if the file was stored, `None` if storage was skipped.

        Raises:
            ValidationError: If the file exceeds the maximum allowed size.
        """
        await self._extract_metadata()

        if not self._should_store_file(node=node):
            return None

        return await self._store_file()

    def delete_file(self) -> None:
        """Delete the uploaded file from the storage backend."""
        if self.storage_id:
            registry.storage.delete(identifier=self.storage_id)



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING, Any

from infrahub_sdk.graphql import Mutation, Query
from infrahub_sdk.types import Order
from prefect import flow

from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.core.constants import InfrahubKind
from infrahub.generators.models import (
    GeneratorDefinitionModel,
    RequestGeneratorRun,
)
from infrahub.services import InfrahubServices  # noqa: TC001  needed for prefect flow
from infrahub.trigger.models import TriggerType
from infrahub.trigger.setup import setup_triggers_specific
from infrahub.workers.dependencies import get_client, get_workflow
from infrahub.workflows.catalogue import REQUEST_GENERATOR_RUN
from infrahub.workflows.utils import add_tags

from .gather import gather_trigger_action_rules
from .models import EventGroupMember  # noqa: TC001  needed for prefect flow

if TYPE_CHECKING:
    from infrahub_sdk.client import InfrahubClient
    from infrahub_sdk.node import InfrahubNode


def get_generator_run_query(definition_id: str, target_ids: list[str]) -> Query:
    return Query(
        name=InfrahubKind.GENERATORDEFINITION,
        query={
            InfrahubKind.GENERATORDEFINITION: {
                "@filters": {
                    "ids": [definition_id],
                },
                "edges": {
                    "node": {
                        "id": None,
                        "name": {
                            "value": None,
                        },
                        "class_name": {
                            "value": None,
                        },
                        "file_path": {
                            "value": None,
                        },
                        "query": {
                            "node": {
                                "name": {
                                    "value": None,
                                },
                            },
                        },
                        "convert_query_response": {
                            "value": None,
                        },
                        "parameters": {
                            "value": None,
                        },
                        "execute_in_proposed_change": {
                            "value": None,
                        },
                        "execute_after_merge": {
                            "value": None,
                        },
                        "targets": {
                            "node": {
                                "id": None,
                                "members": {
                                    "@filters": {
                                        "ids": target_ids,
                                    },
                                    "edges": {
                                        "node": {
                                            "__typename": None,
                                            "id": None,
                                            "display_label": None,
                                        },
                                    },
                                },
                            },
                        },
                        "repository": {
                            "node": {
                                "__typename": None,
                                "id": None,
                                "name": {
                                    "value": None,
                                },
                                f"... on {InfrahubKind.REPOSITORY}": {
                                    "commit": {
                                        "value": None,
                                    },
                                },
                                f"... on {InfrahubKind.READONLYREPOSITORY}": {
                                    "commit": {
                                        "value": None,
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
    )


@flow(
    name="action-add-node-to-group",
    flow_run_name="Adding node={node_id} to group={group_id}",
)
async def add_node_to_group(
    branch_name: str,
    node_id: str,
    group_id: str,
    context: InfrahubContext,  # noqa: ARG001
    service: InfrahubServices,
) -> None:
    await add_tags(branches=[branch_name], nodes=[node_id, group_id])

    mutation = Mutation(
        mutation="RelationshipAdd",
        input_data={"data": {"id": group_id, "name": "members", "nodes": [{"id": node_id}]}},
        query={"ok": None},
    )

    await service.client.execute_graphql(query=mutation.render(), branch_name=branch_name)


@flow(
    name="action-remove-node-from-group",
    flow_run_name="Removing node={node_id} from group={group_id}",
)
async def remove_node_from_group(
    branch_name: str,
    node_id: str,
    group_id: str,
    context: InfrahubContext,  # noqa: ARG001
    service: InfrahubServices,
) -> None:
    await add_tags(branches=[branch_name], nodes=[node_id, group_id])

    mutation = Mutation(
        mutation="RelationshipRemove",
        input_data={"data": {"id": group_id, "name": "members", "nodes": [{"id": node_id}]}},
        query={"ok": None},
    )

    await service.client.execute_graphql(query=mutation.render(), branch_name=branch_name)


@flow(
    name="action-run-generator",
    flow_run_name="Running generator generator_definition_id={generator_definition_id} for nodes={node_ids}",
)
async def run_generator(
    branch_name: str,
    node_ids: list[str],
    generator_definition_id: str,
    context: InfrahubContext,
    service: InfrahubServices,  # noqa: ARG001
) -> None:
    await add_tags(branches=[branch_name], nodes=node_ids + [generator_definition_id])

    client = get_client()

    await _run_generators(
        branch_name=branch_name,
        node_ids=node_ids,
        generator_definition_id=generator_definition_id,
        client=client,
        context=context,
    )


@flow(
    name="action-run-generator-group-event",
    flow_run_name="Running generator",
)
async def run_generator_group_event(
    branch_name: str,
    members: list[EventGroupMember],
    generator_definition_id: str,
    context: InfrahubContext,
    service: InfrahubServices,  # noqa: ARG001
) -> None:
    node_ids = [node.id for node in members]
    await add_tags(branches=[branch_name], nodes=node_ids + [generator_definition_id])

    client = get_client()

    await _run_generators(
        branch_name=branch_name,
        node_ids=node_ids,
        generator_definition_id=generator_definition_id,
        client=client,
        context=context,
    )


@flow(
    name="configure-action-rules",
    flow_run_name="Configure updated action rules and triggers",
)
async def configure_action_rules(
    service: InfrahubServices,
) -> None:
    await setup_triggers_specific(
        gatherer=gather_trigger_action_rules,  # type: ignore[arg-type]
        trigger_type=TriggerType.ACTION_TRIGGER_RULE,
        db=service.database,
    )


async def _get_targets(
    branch_name: str,
    targets: list[dict[str, Any]],
    client: InfrahubClient,
) -> dict[str, dict[str, InfrahubNode]]:
    """Get the targets per kind in order to extract the variables."""

    targets_per_kind: dict[str, dict[str, InfrahubNode]] = defaultdict(dict)

    for target in targets:
        targets_per_kind[target["node"]["__typename"]][target["node"]["id"]] = None

    for kind, values in targets_per_kind.items():
        nodes = await client.filters(
            kind=kind, branch=branch_name, ids=list(values.keys()), populate_store=False, order=Order(disable=True)
        )
        for node in nodes:
            targets_per_kind[kind][node.id] = node

    return targets_per_kind


async def _run_generators(
    branch_name: str,
    node_ids: list[str],
    generator_definition_id: str,
    client: InfrahubClient,
    context: InfrahubContext | None = None,
) -> None:
    """Fetch generator metadata and submit per-target runs.

    Args:
        branch_name: Branch on which to execute.
        node_ids: Node IDs to run against (restricts selection if provided).
        generator_definition_id: Generator definition to execute.
        client: InfrahubClient to query additional data.
        context: Execution context passed to downstream workflow submissions.

    Returns:
        None

    Raises:
        ValueError: If the generator definition is not found or none of the requested
            targets are members of the target group.
    """
    response = await client.execute_graphql(
        query=get_generator_run_query(definition_id=generator_definition_id, target_ids=node_ids).render(),
        branch_name=branch_name,
    )
    if not response[InfrahubKind.GENERATORDEFINITION]["edges"]:
        raise ValueError(f"Generator definition {generator_definition_id} not found")

    data = response[InfrahubKind.GENERATORDEFINITION]["edges"][0]["node"]

    if not data["targets"]["node"]["members"]["edges"]:
        raise ValueError(f"Target {node_ids[0]} is not part of the group {data['targets']['node']['id']}")

    targets = data["targets"]["node"]["members"]["edges"]

    targets_per_kind = await _get_targets(branch_name=branch_name, targets=targets, client=client)

    workflow = get_workflow()

    for target in targets:
        node: InfrahubNode | None = None
        if data["parameters"]["value"]:
            node = targets_per_kind[target["node"]["__typename"]][target["node"]["id"]]

        request_generator_run_model = RequestGeneratorRun(
            generator_definition=GeneratorDefinitionModel(
                definition_id=generator_definition_id,
                definition_name=data["name"]["value"],
                class_name=data["class_name"]["value"],
                file_path=data["file_path"]["value"],
                query_name=data["query"]["node"]["name"]["value"],
                convert_query_response=data["convert_query_response"]["value"],
                group_id=data["targets"]["node"]["id"],
                parameters=data["parameters"]["value"],
                execute_in_proposed_change=data["execute_in_proposed_change"]["value"],
                execute_after_merge=data["execute_after_merge"]["value"],
            ),
            commit=data["repository"]["node"]["commit"]["value"],
            repository_id=data["repository"]["node"]["id"],
            repository_name=data["repository"]["node"]["name"]["value"],
            repository_kind=data["repository"]["node"]["__typename"],
            branch_name=branch_name,
            query=data["query"]["node"]["name"]["value"],
            variables=await node.extract(params=data["parameters"]["value"]) if node else {},
            target_id=target["node"]["id"],
            target_name=target["node"]["display_label"],
        )
        await workflow.submit_workflow(
            workflow=REQUEST_GENERATOR_RUN, context=context, parameters={"model": request_generator_run_model}
        )



from dataclasses import dataclass


@dataclass
class FilterDefinition:
    name: str
    trusted: bool
    source: str


BUILTIN_FILTERS = [
    FilterDefinition(name="abs", trusted=True, source="jinja2"),
    FilterDefinition(name="attr", trusted=False, source="jinja2"),
    FilterDefinition(name="batch", trusted=False, source="jinja2"),
    FilterDefinition(name="capitalize", trusted=True, source="jinja2"),
    FilterDefinition(name="center", trusted=True, source="jinja2"),
    FilterDefinition(name="count", trusted=True, source="jinja2"),
    FilterDefinition(name="d", trusted=True, source="jinja2"),
    FilterDefinition(name="default", trusted=True, source="jinja2"),
    FilterDefinition(name="dictsort", trusted=False, source="jinja2"),
    FilterDefinition(name="e", trusted=True, source="jinja2"),
    FilterDefinition(name="escape", trusted=True, source="jinja2"),
    FilterDefinition(name="filesizeformat", trusted=True, source="jinja2"),
    FilterDefinition(name="first", trusted=True, source="jinja2"),
    FilterDefinition(name="float", trusted=True, source="jinja2"),
    FilterDefinition(name="forceescape", trusted=True, source="jinja2"),
    FilterDefinition(name="format", trusted=True, source="jinja2"),
    FilterDefinition(name="groupby", trusted=False, source="jinja2"),
    FilterDefinition(name="indent", trusted=True, source="jinja2"),
    FilterDefinition(name="int", trusted=True, source="jinja2"),
    FilterDefinition(name="items", trusted=False, source="jinja2"),
    FilterDefinition(name="join", trusted=True, source="jinja2"),
    FilterDefinition(name="last", trusted=True, source="jinja2"),
    FilterDefinition(name="length", trusted=True, source="jinja2"),
    FilterDefinition(name="list", trusted=True, source="jinja2"),
    FilterDefinition(name="lower", trusted=True, source="jinja2"),
    FilterDefinition(name="map", trusted=False, source="jinja2"),
    FilterDefinition(name="max", trusted=True, source="jinja2"),
    FilterDefinition(name="min", trusted=True, source="jinja2"),
    FilterDefinition(name="pprint", trusted=False, source="jinja2"),
    FilterDefinition(name="random", trusted=False, source="jinja2"),
    FilterDefinition(name="reject", trusted=False, source="jinja2"),
    FilterDefinition(name="rejectattr", trusted=False, source="jinja2"),
    FilterDefinition(name="replace", trusted=True, source="jinja2"),
    FilterDefinition(name="reverse", trusted=True, source="jinja2"),
    FilterDefinition(name="round", trusted=True, source="jinja2"),
    FilterDefinition(name="safe", trusted=False, source="jinja2"),
    FilterDefinition(name="select", trusted=False, source="jinja2"),
    FilterDefinition(name="selectattr", trusted=False, source="jinja2"),
    FilterDefinition(name="slice", trusted=True, source="jinja2"),
    FilterDefinition(name="sort", trusted=False, source="jinja2"),
    FilterDefinition(name="string", trusted=True, source="jinja2"),
    FilterDefinition(name="striptags", trusted=True, source="jinja2"),
    FilterDefinition(name="sum", trusted=True, source="jinja2"),
    FilterDefinition(name="title", trusted=True, source="jinja2"),
    FilterDefinition(name="tojson", trusted=False, source="jinja2"),
    FilterDefinition(name="trim", trusted=True, source="jinja2"),
    FilterDefinition(name="truncate", trusted=True, source="jinja2"),
    FilterDefinition(name="unique", trusted=False, source="jinja2"),
    FilterDefinition(name="upper", trusted=True, source="jinja2"),
    FilterDefinition(name="urlencode", trusted=True, source="jinja2"),
    FilterDefinition(name="urlize", trusted=False, source="jinja2"),
    FilterDefinition(name="wordcount", trusted=True, source="jinja2"),
    FilterDefinition(name="wordwrap", trusted=True, source="jinja2"),
    FilterDefinition(name="xmlattr", trusted=False, source="jinja2"),
]


NETUTILS_FILTERS = [
    FilterDefinition(name="abbreviated_interface_name", trusted=True, source="netutils"),
    FilterDefinition(name="abbreviated_interface_name_list", trusted=True, source="netutils"),
    FilterDefinition(name="asn_to_int", trusted=True, source="netutils"),
    FilterDefinition(name="bits_to_name", trusted=True, source="netutils"),
    FilterDefinition(name="bytes_to_name", trusted=True, source="netutils"),
    FilterDefinition(name="canonical_interface_name", trusted=True, source="netutils"),
    FilterDefinition(name="canonical_interface_name_list", trusted=True, source="netutils"),
    FilterDefinition(name="cidr_to_netmask", trusted=True, source="netutils"),
    FilterDefinition(name="cidr_to_netmaskv6", trusted=True, source="netutils"),
    FilterDefinition(name="clean_config", trusted=True, source="netutils"),
    FilterDefinition(name="compare_version_loose", trusted=True, source="netutils"),
    FilterDefinition(name="compare_version_strict", trusted=True, source="netutils"),
    FilterDefinition(name="config_compliance", trusted=True, source="netutils"),
    FilterDefinition(name="config_section_not_parsed", trusted=True, source="netutils"),
    FilterDefinition(name="delimiter_change", trusted=True, source="netutils"),
    FilterDefinition(name="diff_network_config", trusted=True, source="netutils"),
    FilterDefinition(name="feature_compliance", trusted=True, source="netutils"),
    FilterDefinition(name="find_unordered_cfg_lines", trusted=True, source="netutils"),
    FilterDefinition(name="fqdn_to_ip", trusted=False, source="netutils"),
    FilterDefinition(name="get_all_host", trusted=False, source="netutils"),
    FilterDefinition(name="get_broadcast_address", trusted=True, source="netutils"),
    FilterDefinition(name="get_first_usable", trusted=True, source="netutils"),
    FilterDefinition(name="get_ips_sorted", trusted=True, source="netutils"),
    FilterDefinition(name="get_nist_urls", trusted=True, source="netutils"),
    FilterDefinition(name="get_nist_vendor_platform_urls", trusted=True, source="netutils"),
    FilterDefinition(name="get_oui", trusted=True, source="netutils"),
    FilterDefinition(name="get_peer_ip", trusted=True, source="netutils"),
    FilterDefinition(name="get_range_ips", trusted=True, source="netutils"),
    FilterDefinition(name="get_upgrade_path", trusted=True, source="netutils"),
    FilterDefinition(name="get_usable_range", trusted=True, source="netutils"),
    FilterDefinition(name="hash_data", trusted=True, source="netutils"),
    FilterDefinition(name="int_to_asdot", trusted=True, source="netutils"),
    FilterDefinition(name="interface_range_compress", trusted=True, source="netutils"),
    FilterDefinition(name="interface_range_expansion", trusted=True, source="netutils"),
    FilterDefinition(name="ip_addition", trusted=True, source="netutils"),
    FilterDefinition(name="ip_subtract", trusted=True, source="netutils"),
    FilterDefinition(name="ip_to_bin", trusted=True, source="netutils"),
    FilterDefinition(name="ip_to_hex", trusted=True, source="netutils"),
    FilterDefinition(name="ipaddress_address", trusted=True, source="netutils"),
    FilterDefinition(name="ipaddress_interface", trusted=True, source="netutils"),
    FilterDefinition(name="ipaddress_network", trusted=True, source="netutils"),
    FilterDefinition(name="is_classful", trusted=True, source="netutils"),
    FilterDefinition(name="is_fqdn_resolvable", trusted=False, source="netutils"),
    FilterDefinition(name="is_ip", trusted=True, source="netutils"),
    FilterDefinition(name="is_ip_range", trusted=True, source="netutils"),
    FilterDefinition(name="is_ip_within", trusted=True, source="netutils"),
    FilterDefinition(name="is_netmask", trusted=True, source="netutils"),
    FilterDefinition(name="is_network", trusted=True, source="netutils"),
    FilterDefinition(name="is_reversible_wildcardmask", trusted=True, source="netutils"),
    FilterDefinition(name="is_valid_mac", trusted=True, source="netutils"),
    FilterDefinition(name="longest_prefix_match", trusted=True, source="netutils"),
    FilterDefinition(name="mac_normalize", trusted=True, source="netutils"),
    FilterDefinition(name="mac_to_format", trusted=True, source="netutils"),
    FilterDefinition(name="mac_to_int", trusted=True, source="netutils"),
    FilterDefinition(name="mac_type", trusted=True, source="netutils"),
    FilterDefinition(name="name_to_bits", trusted=True, source="netutils"),
    FilterDefinition(name="name_to_bytes", trusted=True, source="netutils"),
    FilterDefinition(name="name_to_name", trusted=True, source="netutils"),
    FilterDefinition(name="netmask_to_cidr", trusted=True, source="netutils"),
    FilterDefinition(name="netmask_to_wildcardmask", trusted=True, source="netutils"),
    FilterDefinition(name="normalise_delimiter_caret_c", trusted=True, source="netutils"),
    FilterDefinition(name="paloalto_panos_brace_to_set", trusted=True, source="netutils"),
    FilterDefinition(name="paloalto_panos_clean_newlines", trusted=True, source="netutils"),
    FilterDefinition(name="regex_findall", trusted=False, source="netutils"),
    FilterDefinition(name="regex_match", trusted=False, source="netutils"),
    FilterDefinition(name="regex_search", trusted=False, source="netutils"),
    FilterDefinition(name="regex_split", trusted=False, source="netutils"),
    FilterDefinition(name="regex_sub", trusted=False, source="netutils"),
    FilterDefinition(name="sanitize_config", trusted=True, source="netutils"),
    FilterDefinition(name="section_config", trusted=True, source="netutils"),
    FilterDefinition(name="sort_interface_list", trusted=True, source="netutils"),
    FilterDefinition(name="split_interface", trusted=True, source="netutils"),
    FilterDefinition(name="uptime_seconds_to_string", trusted=True, source="netutils"),
    FilterDefinition(name="uptime_string_to_seconds", trusted=True, source="netutils"),
    FilterDefinition(name="version_metadata", trusted=True, source="netutils"),
    FilterDefinition(name="vlanconfig_to_list", trusted=True, source="netutils"),
    FilterDefinition(name="vlanlist_to_config", trusted=True, source="netutils"),
    FilterDefinition(name="wildcardmask_to_netmask", trusted=True, source="netutils"),
]


AVAILABLE_FILTERS = BUILTIN_FILTERS + NETUTILS_FILTERS
